#!/usr/bin/python
# Copyright (c) 2017, Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: ipa_user
author: Thomas Krahn (@Nosmoht)
short_description: Manage FreeIPA users
description:
  - Add, modify and delete user within IPA server.
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  displayname:
    description: Display name.
    type: str
  update_password:
    description:
      - Set password for a user.
    type: str
    default: 'always'
    choices: [always, on_create]
  givenname:
    description:
      - First name.
      - If user does not exist and O(state=present), the usage of O(givenname) is required.
    type: str
  krbpasswordexpiration:
    description:
      - Date at which the user password expires.
      - In the format YYYYMMddHHmmss.
      - For example V(20180121182022) expires on 21 January 2018 at 18:20:22.
    type: str
  loginshell:
    description: Login shell.
    type: str
  mail:
    description:
      - List of mail addresses assigned to the user.
      - If an empty list is passed all assigned email addresses are deleted.
      - If None is passed email addresses are not checked nor changed.
    type: list
    elements: str
  password:
    description:
      - Password for a user.
      - It is not set for an existing user unless O(update_password=always), which is the default.
    type: str
  sn:
    description:
      - Surname.
      - If user does not exist and O(state=present), the usage of O(sn) is required.
    type: str
  sshpubkey:
    description:
      - List of public SSH key.
      - If an empty list is passed all assigned public keys are deleted.
      - If None is passed SSH public keys are not checked nor changed.
    type: list
    elements: str
  state:
    description: State to ensure.
    default: "present"
    choices: ["absent", "disabled", "enabled", "present"]
    type: str
  telephonenumber:
    description:
      - List of telephone numbers assigned to the user.
      - If an empty list is passed all assigned telephone numbers are deleted.
      - If None is passed telephone numbers are not checked nor changed.
    type: list
    elements: str
  title:
    description: Title.
    type: str
  uid:
    description: Uid of the user.
    required: true
    aliases: ["name"]
    type: str
  uidnumber:
    description:
      - Account Settings UID/Posix User ID number.
    type: str
  gidnumber:
    description:
      - Posix Group ID.
    type: str
  homedirectory:
    description:
      - Default home directory of the user.
    type: str
    version_added: '0.2.0'
  userauthtype:
    description:
      - The authentication type to use for the user.
      - To remove all authentication types from the user, use an empty list V([]).
      - The choice V(idp) and V(passkey) has been added in community.general 8.1.0.
    choices: ["password", "radius", "otp", "pkinit", "hardened", "idp", "passkey"]
    type: list
    elements: str
    version_added: '1.2.0'
extends_documentation_fragment:
  - community.general._ipa.documentation
  - community.general._ipa.connection_notes
  - community.general._attributes

requirements:
  - base64
  - hashlib
"""

EXAMPLES = r"""
- name: Ensure pinky is present and always reset password
  community.general.ipa_user:
    name: pinky
    state: present
    krbpasswordexpiration: 20200119235959
    givenname: Pinky
    sn: Acme
    mail:
      - pinky@acme.com
    telephonenumber:
      - '+555123456'
    sshpubkey:
      - ssh-rsa ....
      - ssh-dsa ....
    uidnumber: '1001'
    gidnumber: '100'
    homedirectory: /home/pinky
    ipa_host: ipa.example.com
    ipa_user: admin
    ipa_pass: topsecret

- name: Ensure brain is absent
  community.general.ipa_user:
    name: brain
    state: absent
    ipa_host: ipa.example.com
    ipa_user: admin
    ipa_pass: topsecret

- name: Ensure pinky is present but don't reset password if already exists
  community.general.ipa_user:
    name: pinky
    state: present
    givenname: Pinky
    sn: Acme
    password: zounds
    ipa_host: ipa.example.com
    ipa_user: admin
    ipa_pass: topsecret
    update_password: on_create

- name: Ensure pinky is present and using one time password and RADIUS authentication
  community.general.ipa_user:
    name: pinky
    state: present
    userauthtype:
      - otp
      - radius
    ipa_host: ipa.example.com
    ipa_user: admin
    ipa_pass: topsecret
"""

RETURN = r"""
user:
  description: User as returned by IPA API.
  returned: always
  type: dict
"""

import base64
import hashlib
import traceback

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._ipa import IPAClient, ipa_argument_spec


class UserIPAClient(IPAClient):
    def __init__(self, module, host, port, protocol):
        super().__init__(module, host, port, protocol)

    def user_find(self, name):
        return self._post_json(method="user_find", name=None, item={"all": True, "uid": name})

    def user_add(self, name, item):
        return self._post_json(method="user_add", name=name, item=item)

    def user_mod(self, name, item):
        return self._post_json(method="user_mod", name=name, item=item)

    def user_del(self, name):
        return self._post_json(method="user_del", name=name)

    def user_disable(self, name):
        return self._post_json(method="user_disable", name=name)

    def user_enable(self, name):
        return self._post_json(method="user_enable", name=name)


def get_user_dict(
    displayname=None,
    givenname=None,
    krbpasswordexpiration=None,
    loginshell=None,
    mail=None,
    nsaccountlock=False,
    sn=None,
    sshpubkey=None,
    telephonenumber=None,
    title=None,
    userpassword=None,
    gidnumber=None,
    uidnumber=None,
    homedirectory=None,
    userauthtype=None,
):
    user = {}
    if displayname is not None:
        user["displayname"] = displayname
    if krbpasswordexpiration is not None:
        user["krbpasswordexpiration"] = f"{krbpasswordexpiration}Z"
    if givenname is not None:
        user["givenname"] = givenname
    if loginshell is not None:
        user["loginshell"] = loginshell
    if mail is not None:
        user["mail"] = mail
    user["nsaccountlock"] = nsaccountlock
    if sn is not None:
        user["sn"] = sn
    if sshpubkey is not None:
        user["ipasshpubkey"] = sshpubkey
    if telephonenumber is not None:
        user["telephonenumber"] = telephonenumber
    if title is not None:
        user["title"] = title
    if userpassword is not None:
        user["userpassword"] = userpassword
    if gidnumber is not None:
        user["gidnumber"] = gidnumber
    if uidnumber is not None:
        user["uidnumber"] = uidnumber
    if homedirectory is not None:
        user["homedirectory"] = homedirectory
    if userauthtype is not None:
        user["ipauserauthtype"] = userauthtype

    return user


def get_user_diff(client, ipa_user, module_user):
    """
        Return the keys of each dict whereas values are different. Unfortunately the IPA
        API returns everything as a list even if only a single value is possible.
        Therefore some more complexity is needed.
        The method will check if the value type of module_user.attr is not a list and
        create a list with that element if the same attribute in ipa_user is list. In this way I hope that the method
        must not be changed if the returned API dict is changed.
    :param ipa_user:
    :param module_user:
    :return:
    """
    # sshpubkeyfp is the list of ssh key fingerprints. IPA doesn't return the keys itself but instead the fingerprints.
    # These are used for comparison.
    sshpubkey = None
    if "ipasshpubkey" in module_user:
        hash_algo = "md5"
        if "sshpubkeyfp" in ipa_user and ipa_user["sshpubkeyfp"][0][:7].upper() == "SHA256:":
            hash_algo = "sha256"
        module_user["sshpubkeyfp"] = [
            get_ssh_key_fingerprint(pubkey, hash_algo) for pubkey in module_user["ipasshpubkey"]
        ]
        # Remove the ipasshpubkey element as it is not returned from IPA but save its value to be used later on
        sshpubkey = module_user["ipasshpubkey"]
        del module_user["ipasshpubkey"]

    result = client.get_diff(ipa_data=ipa_user, module_data=module_user)

    # If there are public keys, remove the fingerprints and add them back to the dict
    if sshpubkey is not None:
        del module_user["sshpubkeyfp"]
        module_user["ipasshpubkey"] = sshpubkey
    return result


def get_ssh_key_fingerprint(ssh_key, hash_algo="sha256"):
    """
    Return the public key fingerprint of a given public SSH key
    in format "[fp] [comment] (ssh-rsa)" where fp is of the format:
    FB:0C:AC:0A:07:94:5B:CE:75:6E:63:32:13:AD:AD:D7
    for md5 or
    SHA256:[base64]
    for sha256
    Comments are assumed to be all characters past the second
    whitespace character in the sshpubkey string.
    :param ssh_key:
    :param hash_algo:
    :return:
    """
    parts = ssh_key.strip().split(None, 2)
    if len(parts) == 0:
        return None
    key_type = parts[0]
    key = base64.b64decode(parts[1].encode("ascii"))

    if hash_algo == "md5":
        fp_plain = hashlib.md5(key).hexdigest()
        key_fp = ":".join(a + b for a, b in zip(fp_plain[::2], fp_plain[1::2])).upper()
    elif hash_algo == "sha256":
        fp_plain = base64.b64encode(hashlib.sha256(key).digest()).decode("ascii").rstrip("=")
        key_fp = f"SHA256:{fp_plain}"
    if len(parts) < 3:
        return f"{key_fp} ({key_type})"
    else:
        comment = parts[2]
        return f"{key_fp} {comment} ({key_type})"


def ensure(module, client):
    state = module.params["state"]
    name = module.params["uid"]
    nsaccountlock = state == "disabled"

    module_user = get_user_dict(
        displayname=module.params.get("displayname"),
        krbpasswordexpiration=module.params.get("krbpasswordexpiration"),
        givenname=module.params.get("givenname"),
        loginshell=module.params["loginshell"],
        mail=module.params["mail"],
        sn=module.params["sn"],
        sshpubkey=module.params["sshpubkey"],
        nsaccountlock=nsaccountlock,
        telephonenumber=module.params["telephonenumber"],
        title=module.params["title"],
        userpassword=module.params["password"],
        gidnumber=module.params.get("gidnumber"),
        uidnumber=module.params.get("uidnumber"),
        homedirectory=module.params.get("homedirectory"),
        userauthtype=module.params.get("userauthtype"),
    )

    update_password = module.params.get("update_password")
    ipa_user = client.user_find(name=name)

    changed = False
    if state in ["present", "enabled", "disabled"]:
        if not ipa_user:
            changed = True
            if not module.check_mode:
                ipa_user = client.user_add(name=name, item=module_user)
        else:
            if update_password == "on_create":
                module_user.pop("userpassword", None)
            diff = get_user_diff(client, ipa_user, module_user)
            if len(diff) > 0:
                changed = True
                if not module.check_mode:
                    ipa_user = client.user_mod(name=name, item=module_user)
    else:
        if ipa_user:
            changed = True
            if not module.check_mode:
                client.user_del(name)

    return changed, ipa_user


def main():
    argument_spec = ipa_argument_spec()
    argument_spec.update(
        displayname=dict(type="str"),
        givenname=dict(type="str"),
        update_password=dict(type="str", default="always", choices=["always", "on_create"], no_log=False),
        krbpasswordexpiration=dict(type="str", no_log=False),
        loginshell=dict(type="str"),
        mail=dict(type="list", elements="str"),
        sn=dict(type="str"),
        uid=dict(type="str", required=True, aliases=["name"]),
        gidnumber=dict(type="str"),
        uidnumber=dict(type="str"),
        password=dict(type="str", no_log=True),
        sshpubkey=dict(type="list", elements="str"),
        state=dict(type="str", default="present", choices=["present", "absent", "enabled", "disabled"]),
        telephonenumber=dict(type="list", elements="str"),
        title=dict(type="str"),
        homedirectory=dict(type="str"),
        userauthtype=dict(
            type="list", elements="str", choices=["password", "radius", "otp", "pkinit", "hardened", "idp", "passkey"]
        ),
    )

    module = AnsibleModule(argument_spec=argument_spec, supports_check_mode=True)

    client = UserIPAClient(
        module=module,
        host=module.params["ipa_host"],
        port=module.params["ipa_port"],
        protocol=module.params["ipa_prot"],
    )

    # If sshpubkey is defined as None than module.params['sshpubkey'] is [None]. IPA itself returns None (not a list).
    # Therefore a small check here to replace list(None) by None. Otherwise get_user_diff() would return sshpubkey
    # as different which should be avoided.
    if module.params["sshpubkey"] is not None:
        if len(module.params["sshpubkey"]) == 1 and module.params["sshpubkey"][0] == "":
            module.params["sshpubkey"] = None

    try:
        client.login(username=module.params["ipa_user"], password=module.params["ipa_pass"])
        changed, user = ensure(module, client)
        module.exit_json(changed=changed, user=user)
    except Exception as e:
        module.fail_json(msg=f"{e}", exception=traceback.format_exc())


if __name__ == "__main__":
    main()



# Copyright (c) 2017, Patrick Deelman <patrick@patrickdeelman.nl>
# Copyright (c) 2017 Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
from __future__ import annotations

DOCUMENTATION = r"""
name: passwordstore
author:
  - Patrick Deelman (!UNKNOWN) <patrick@patrickdeelman.nl>
short_description: Manage passwords with passwordstore.org's pass utility
description:
  - Enables Ansible to retrieve, create or update passwords from the passwordstore.org pass utility. It can also retrieve,
    create or update YAML style keys stored as multilines in the passwordfile.
  - To avoid problems when accessing multiple secrets at once, add C(auto-expand-secmem) to C(~/.gnupg/gpg-agent.conf). Where
    this is not possible, consider using O(lock=readwrite) instead.
options:
  _terms:
    description: Query key.
    required: true
  directory:
    description:
      - The directory of the password store.
      - If O(backend=pass), the default is V(~/.password-store) is used.
      - If O(backend=gopass), then the default is the C(path) field in C(~/.config/gopass/config.yml), falling back to V(~/.local/share/gopass/stores/root)
        if C(path) is not defined in the gopass config.
    type: path
    vars:
      - name: passwordstore
    env:
      - name: PASSWORD_STORE_DIR
  create:
    description: Create the password or the subkey if it does not already exist. Takes precedence over O(missing).
    type: bool
    default: false
  overwrite:
    description: Overwrite the password or the subkey if it does already exist.
    type: bool
    default: false
  umask:
    description:
      - Sets the umask for the created V(.gpg) files. The first octed must be greater than 3 (user readable).
      - Note pass' default value is V('077').
    type: string
    env:
      - name: PASSWORD_STORE_UMASK
    version_added: 1.3.0
  returnall:
    description: Return all the content of the password, not only the first line.
    type: bool
    default: false
  subkey:
    description:
      - By default return a specific subkey of the password. When set to V(password), always returns the first line.
      - With O(overwrite=true), it creates the subkey and returns it.
    type: str
    default: password
  userpass:
    description: Specify a password to save, instead of a generated one.
    type: str
  length:
    description: The length of the generated password.
    type: integer
    default: 16
  backup:
    description: Used with O(overwrite=true). Backup the previous password or subkey in a subkey.
    type: bool
    default: false
  nosymbols:
    description: Use alphanumeric characters.
    type: bool
    default: false
  missing:
    description:
      - List of preference about what to do if the password file is missing.
      - If O(create=true), the value for this option is ignored and assumed to be V(create).
      - If set to V(error), the lookup fails out if the passname does not exist.
      - If set to V(create), the passname is created with the provided length O(length) if it does not exist.
      - If set to V(empty) or V(warn), it returns a V(none) in case the passname does not exist. When using C(lookup) and
        not C(query), this is translated to an empty string.
    version_added: 3.1.0
    type: str
    default: error
    choices:
      - error
      - warn
      - empty
      - create
  lock:
    description:
      - How to synchronize operations.
      - The default of V(write) only synchronizes write operations.
      - V(readwrite) synchronizes all operations (including read). This makes sure that gpg-agent is never called in parallel.
      - V(none) does not do any synchronization.
    ini:
      - section: passwordstore_lookup
        key: lock
    type: str
    default: write
    choices:
      - readwrite
      - write
      - none
    version_added: 4.5.0
  locktimeout:
    description:
      - Lock timeout applied when O(lock) is not V(none).
      - Time with a unit suffix, V(s), V(m), V(h) for seconds, minutes, and hours, respectively. For example, V(900s) equals
        V(15m).
      - Correlates with C(pinentry-timeout) in C(~/.gnupg/gpg-agent.conf), see C(man gpg-agent) for details.
    ini:
      - section: passwordstore_lookup
        key: locktimeout
    type: str
    default: 15m
    version_added: 4.5.0
  backend:
    description:
      - Specify which backend to use.
      - Defaults to V(pass), passwordstore.org's original pass utility.
      - V(gopass) support is incomplete.
    ini:
      - section: passwordstore_lookup
        key: backend
    vars:
      - name: passwordstore_backend
    type: str
    default: pass
    choices:
      - pass
      - gopass
    version_added: 5.2.0
  timestamp:
    description: Add the password generation information to the end of the file.
    type: bool
    default: true
    version_added: 8.1.0
  preserve:
    description: Include the old (edited) password inside the pass file.
    type: bool
    default: true
    version_added: 8.1.0
  missing_subkey:
    description:
      - Preference about what to do if the password subkey is missing.
      - If set to V(error), the lookup fails out if the subkey does not exist.
      - If set to V(empty) or V(warn), it returns a V(none) in case the subkey does not exist.
    version_added: 8.6.0
    type: str
    default: empty
    choices:
      - error
      - warn
      - empty
    ini:
      - section: passwordstore_lookup
        key: missing_subkey
notes:
  - The lookup supports passing all options as lookup parameters since community.general 6.0.0.
"""
EXAMPLES = r"""
ansible.cfg: |
  [passwordstore_lookup]
  lock=readwrite
  locktimeout=45s
  missing_subkey=warn

tasks.yml: |-
  ---

  # Debug is used for examples, BAD IDEA to show passwords on screen
  - name: Basic lookup. Fails if example/test does not exist
    ansible.builtin.debug:
      msg: "{{ lookup('community.general.passwordstore', 'example/test')}}"

  - name: Basic lookup. Warns if example/test does not exist and returns empty string
    ansible.builtin.debug:
      msg: "{{ lookup('community.general.passwordstore', 'example/test', missing='warn')}}"

  - name: Create pass with random 16 character password. If password exists just give the password
    ansible.builtin.debug:
      var: mypassword
    vars:
      mypassword: "{{ lookup('community.general.passwordstore', 'example/test', create=true)}}"

  - name: Create pass with random 16 character password. If password exists just give the password
    ansible.builtin.debug:
      var: mypassword
    vars:
      mypassword: "{{ lookup('community.general.passwordstore', 'example/test', missing='create')}}"

  - name: >-
      Create a random 16 character password in a subkey. If the password file already exists, just add the subkey in it.
      If the subkey exists, returns it
    ansible.builtin.debug:
      msg: "{{ lookup('community.general.passwordstore', 'example/test', create=true, subkey='foo') }}"

  - name: >-
      Create a random 16 character password in a subkey. Overwrite if it already exists and backup the old one.
    ansible.builtin.debug:
      msg: "{{ lookup('community.general.passwordstore', 'example/test', create=true, subkey='user', overwrite=true, backup=true) }}"

  - name: Prints 'abc' if example/test does not exist, just give the password otherwise
    ansible.builtin.debug:
      var: mypassword
    vars:
      mypassword: >-
        {{ lookup('community.general.passwordstore', 'example/test', missing='empty')
           | default('abc', true) }}

  - name: Different size password
    ansible.builtin.debug:
      msg: "{{ lookup('community.general.passwordstore', 'example/test', create=true, length=42)}}"

  - name: >-
      Create password and overwrite the password if it exists.
      As a bonus, this module includes the old password inside the pass file
    ansible.builtin.debug:
      msg: "{{ lookup('community.general.passwordstore', 'example/test', create=true, overwrite=true)}}"

  - name: Create an alphanumeric password
    ansible.builtin.debug:
      msg: "{{ lookup('community.general.passwordstore', 'example/test', create=true, nosymbols=true) }}"

  - name: Return the value for user in the KV pair user, username
    ansible.builtin.debug:
      msg: "{{ lookup('community.general.passwordstore', 'example/test', subkey='user')}}"

  - name: Return the entire password file content
    ansible.builtin.set_fact:
      passfilecontent: "{{ lookup('community.general.passwordstore', 'example/test', returnall=true)}}"
"""

RETURN = r"""
_raw:
  description:
    - A password.
  type: list
  elements: str
"""

import os
import re
import subprocess
import time
from contextlib import contextmanager

import yaml
from ansible import constants as C
from ansible.errors import AnsibleAssertionError, AnsibleError
from ansible.module_utils.common.text.converters import to_bytes, to_native, to_text
from ansible.module_utils.parsing.convert_bool import boolean
from ansible.plugins.lookup import LookupBase
from ansible.utils.display import Display
from ansible.utils.encrypt import random_password

from ansible_collections.community.general.plugins.module_utils._filelock import FileLock
from ansible_collections.community.general.plugins.plugin_utils._lookup import check_for_wrong_terms

display = Display()


def run_backend_cmd(cmd, *, input=None, env=None):
    result = subprocess.run(
        cmd,
        check=False,
        capture_output=True,
        input=to_bytes(input, errors="surrogate_or_strict") if input else None,
        env=env,
    )
    b_out, b_err = result.stdout, result.stderr
    retcode = result.returncode
    if retcode == 0 and (
        b"encryption failed: Unusable public key" in b_out or b"encryption failed: Unusable public key" in b_err
    ):
        retcode = 78  # os.EX_CONFIG
    if retcode != 0:
        raise subprocess.CalledProcessError(retcode, cmd, to_native(b_out + b_err, errors="surrogate_or_strict"))
    return b_out


class LookupModule(LookupBase):
    def __init__(self, loader=None, templar=None, **kwargs):
        super().__init__(loader, templar, **kwargs)
        self.realpass = None

    def is_real_pass(self):
        if self.realpass is None:
            try:
                passoutput = to_text(
                    run_backend_cmd([self.pass_cmd, "--version"], env=self.env), errors="surrogate_or_strict"
                )
                self.realpass = "pass: the standard unix password manager" in passoutput
            except subprocess.CalledProcessError as e:
                raise AnsibleError(f"exit code {e.returncode} while running {e.cmd}. Error output: {e.output}") from e

        return self.realpass

    def parse_params(self, term):
        # I went with the "traditional" param followed with space separated KV pairs.
        # Waiting for final implementation of lookup parameter parsing.
        # See: https://github.com/ansible/ansible/issues/12255
        params = term.split()
        if len(params) > 0:
            # the first param is the pass-name
            self.passname = params[0]
            # next parse the optional parameters in keyvalue pairs
            try:
                for param in params[1:]:
                    name, value = param.split("=", 1)
                    if name not in self.paramvals:
                        raise AnsibleAssertionError(f"{name} not in paramvals")
                    self.paramvals[name] = value
            except (ValueError, AssertionError) as e:
                raise AnsibleError(e) from e
            # check and convert values
            try:
                for key in ["create", "returnall", "overwrite", "backup", "nosymbols"]:
                    if not isinstance(self.paramvals[key], bool):
                        self.paramvals[key] = boolean(self.paramvals[key])
            except (ValueError, AssertionError) as e:
                raise AnsibleError(e) from e
            if self.paramvals["missing"] not in ["error", "warn", "create", "empty"]:
                raise AnsibleError(f"{self.paramvals['missing']} is not a valid option for missing")
            if not isinstance(self.paramvals["length"], int):
                if self.paramvals["length"].isdigit():
                    self.paramvals["length"] = int(self.paramvals["length"])
                else:
                    raise AnsibleError(f"{self.paramvals['length']} is not a correct value for length")

            if self.paramvals["create"]:
                self.paramvals["missing"] = "create"

            # Collect pass environment variables from the plugin's parameters.
            self.env = os.environ.copy()
            self.env["LANGUAGE"] = "C"  # make sure to get errors in English as required by run_backend_cmd

            if self.backend == "gopass":
                self.env["GOPASS_NO_REMINDER"] = "YES"
            elif os.path.isdir(self.paramvals["directory"]):
                # Set PASSWORD_STORE_DIR
                self.env["PASSWORD_STORE_DIR"] = self.paramvals["directory"]
            elif self.is_real_pass():
                raise AnsibleError(f"Passwordstore directory '{self.paramvals['directory']}' does not exist")

            # Set PASSWORD_STORE_UMASK if umask is set
            if self.paramvals.get("umask") is not None:
                if len(self.paramvals["umask"]) != 3:
                    raise AnsibleError("Passwordstore umask must have a length of 3.")
                elif int(self.paramvals["umask"][0]) > 3:
                    raise AnsibleError("Passwordstore umask not allowed (password not user readable).")
                else:
                    self.env["PASSWORD_STORE_UMASK"] = self.paramvals["umask"]

    def check_pass(self):
        try:
            self.passoutput = to_text(
                run_backend_cmd([self.pass_cmd, "show"] + [self.passname], env=self.env), errors="surrogate_or_strict"
            ).splitlines()
            self.password = self.passoutput[0]
            self.passdict = {}
            try:
                values = yaml.safe_load("\n".join(self.passoutput[1:]))
                for key, item in values.items():
                    self.passdict[key] = item
            except (yaml.YAMLError, AttributeError):
                for line in self.passoutput[1:]:
                    if ":" in line:
                        name, value = line.split(":", 1)
                        self.passdict[name.strip()] = value.strip()
            if (
                self.backend == "gopass"
                or os.path.isfile(os.path.join(self.paramvals["directory"], f"{self.passname}.gpg"))
                or not self.is_real_pass()
            ):
                # When using real pass, only accept password as found if there is a .gpg file for it (might be a tree node otherwise)
                return True
        except subprocess.CalledProcessError as e:
            # 'not in password store' is the expected error if a password wasn't found
            if "not in the password store" not in e.output:
                raise AnsibleError(f"exit code {e.returncode} while running {e.cmd}. Error output: {e.output}") from e

        if self.paramvals["missing"] == "error":
            raise AnsibleError(f"passwordstore: passname {self.passname} not found and missing=error is set")
        elif self.paramvals["missing"] == "warn":
            display.warning(f"passwordstore: passname {self.passname} not found")

        return False

    def get_newpass(self):
        if self.paramvals["nosymbols"]:
            chars = C.DEFAULT_PASSWORD_CHARS[:62]
        else:
            chars = C.DEFAULT_PASSWORD_CHARS

        if self.paramvals["userpass"]:
            newpass = self.paramvals["userpass"]
        else:
            newpass = random_password(length=self.paramvals["length"], chars=chars)
        return newpass

    def update_password(self):
        # generate new password, insert old lines from current result and return new password
        # if the target is a subkey, only modify the subkey
        newpass = self.get_newpass()
        datetime = time.strftime("%d/%m/%Y %H:%M:%S")
        subkey = self.paramvals["subkey"]

        if subkey != "password":
            msg_lines = []
            subkey_exists = False
            subkey_line = f"{subkey}: {newpass}"
            oldpass = None

            for line in self.passoutput:
                if line.startswith(f"{subkey}: "):
                    oldpass = self.passdict[subkey]
                    line = subkey_line
                    subkey_exists = True

                msg_lines.append(line)

            if not subkey_exists:
                msg_lines.insert(2, subkey_line)

            if self.paramvals["timestamp"] and self.paramvals["backup"] and oldpass and oldpass != newpass:
                msg_lines.append(f"lookup_pass: old subkey '{subkey}' password was {oldpass} (Updated on {datetime})\n")

            msg = os.linesep.join(msg_lines)

        else:
            msg = newpass

            if self.paramvals["preserve"] or self.paramvals["timestamp"]:
                msg += "\n"
                if self.paramvals["preserve"] and self.passoutput[1:]:
                    msg += "\n".join(self.passoutput[1:])
                    msg += "\n"
                if self.paramvals["timestamp"] and self.paramvals["backup"]:
                    msg += f"lookup_pass: old password was {self.password} (Updated on {datetime})\n"

        try:
            run_backend_cmd([self.pass_cmd, "insert", "-f", "-m", self.passname], input=msg, env=self.env)
        except subprocess.CalledProcessError as e:
            raise AnsibleError(f"exit code {e.returncode} while running {e.cmd}. Error output: {e.output}") from e
        return newpass

    def generate_password(self):
        # generate new file and insert lookup_pass: Generated by Ansible on {date}
        # use pwgen to generate the password and insert values with pass -m
        newpass = self.get_newpass()
        datetime = time.strftime("%d/%m/%Y %H:%M:%S")
        subkey = self.paramvals["subkey"]

        if subkey != "password":
            msg = f"\n\n{subkey}: {newpass}"
        else:
            msg = newpass

        if self.paramvals["timestamp"]:
            msg += f"\nlookup_pass: First generated by ansible on {datetime}\n"

        try:
            run_backend_cmd([self.pass_cmd, "insert", "-f", "-m", self.passname], input=msg, env=self.env)
        except subprocess.CalledProcessError as e:
            raise AnsibleError(f"exit code {e.returncode} while running {e.cmd}. Error output: {e.output}") from e

        return newpass

    def get_passresult(self):
        if self.paramvals["returnall"]:
            return os.linesep.join(self.passoutput)
        if self.paramvals["subkey"] == "password":
            return self.password
        else:
            if self.paramvals["subkey"] in self.passdict:
                return self.passdict[self.paramvals["subkey"]]
            else:
                if self.paramvals["missing_subkey"] == "error":
                    raise AnsibleError(
                        f"passwordstore: subkey {self.paramvals['subkey']} for passname {self.passname} not found and missing_subkey=error is set"
                    )

                if self.paramvals["missing_subkey"] == "warn":
                    display.warning(
                        f"passwordstore: subkey {self.paramvals['subkey']} for passname {self.passname} not found"
                    )

                return None

    @contextmanager
    def opt_lock(self, type):
        if self.get_option("lock") == type:
            tmpdir = os.environ.get("TMPDIR", "/tmp")
            user = os.environ.get("USER")
            lockfile = os.path.join(tmpdir, f".{user}.passwordstore.lock")
            with FileLock().lock_file(lockfile, tmpdir, self.lock_timeout):
                self.locked = type
                yield
            self.locked = None
        else:
            yield

    def setup(self, variables):
        self.backend = self.get_option("backend")
        self.pass_cmd = self.backend  # pass and gopass are commands as well
        self.locked = None
        timeout = self.get_option("locktimeout")
        if not re.match("^[0-9]+[smh]$", timeout):
            raise AnsibleError(f"{timeout} is not a correct value for locktimeout")
        unit_to_seconds = {"s": 1, "m": 60, "h": 3600}
        self.lock_timeout = int(timeout[:-1]) * unit_to_seconds[timeout[-1]]

        directory = self.get_option("directory")
        if directory is None:
            if self.backend == "gopass":
                try:
                    with open(os.path.expanduser("~/.config/gopass/config.yml")) as f:
                        directory = yaml.safe_load(f)["path"]
                except (FileNotFoundError, KeyError, yaml.YAMLError):
                    directory = os.path.expanduser("~/.local/share/gopass/stores/root")
            else:
                directory = os.path.expanduser("~/.password-store")

        self.paramvals = {
            "subkey": self.get_option("subkey"),
            "directory": directory,
            "create": self.get_option("create"),
            "returnall": self.get_option("returnall"),
            "overwrite": self.get_option("overwrite"),
            "nosymbols": self.get_option("nosymbols"),
            "userpass": self.get_option("userpass") or "",
            "length": self.get_option("length"),
            "backup": self.get_option("backup"),
            "missing": self.get_option("missing"),
            "umask": self.get_option("umask"),
            "timestamp": self.get_option("timestamp"),
            "preserve": self.get_option("preserve"),
            "missing_subkey": self.get_option("missing_subkey"),
        }

    def run(self, terms, variables, **kwargs):
        self.set_options(var_options=variables, direct=kwargs)
        check_for_wrong_terms(self, direct=kwargs)
        self.setup(variables)
        result = []

        for term in terms:
            self.parse_params(term)  # parse the input into paramvals
            with self.opt_lock("readwrite"):
                if self.check_pass():  # password file exists
                    if self.paramvals["overwrite"]:  # if "overwrite", always update password
                        with self.opt_lock("write"):
                            result.append(self.update_password())
                    elif (
                        self.paramvals["subkey"] != "password"
                        and not self.passdict.get(self.paramvals["subkey"])
                        and self.paramvals["missing"] == "create"
                    ):  # target is a subkey, this subkey is not in passdict BUT missing == create
                        with self.opt_lock("write"):
                            result.append(self.update_password())
                    else:
                        result.append(self.get_passresult())
                else:  # password does not exist
                    if self.paramvals["missing"] == "create":
                        with self.opt_lock("write"):
                            if (
                                self.locked == "write" and self.check_pass()
                            ):  # lookup password again if under write lock
                                result.append(self.get_passresult())
                            else:
                                result.append(self.generate_password())
                    else:
                        result.append(None)

        return result



# Copyright (c) 2020, Adam Migus <adam@migus.org>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
from __future__ import annotations

DOCUMENTATION = r"""
name: tss
author: Adam Migus (@amigus) <adam@migus.org>
short_description: Get secrets from Delinea Secret Server
version_added: 1.0.0
description:
  - Uses the Thycotic Secret Server Python SDK to get Secrets from Secret Server using token authentication with O(username)
    and O(password) on the REST API at O(base_url).
  - When using self-signed certificates the environment variable E(REQUESTS_CA_BUNDLE) can be set to a file containing the
    trusted certificates (in C(.pem) format).
  - For example, C(export REQUESTS_CA_BUNDLE='/etc/ssl/certs/ca-bundle.trust.crt').
requirements:
  - python-tss-sdk - https://pypi.org/project/python-tss-sdk/
options:
  _terms:
    description: The integer ID of the secret.
    required: true
    type: list
    elements: int
  secret_path:
    description: Indicate a full path of secret including folder and secret name when the secret ID is set to 0.
    required: false
    type: str
    version_added: 7.2.0
  fetch_secret_ids_from_folder:
    description:
      - Boolean flag which indicates whether secret IDs are in a folder is fetched by folder ID or not.
      - V(true) then the terms are considered as a folder IDs. Otherwise (default), they are considered as secret IDs.
    required: false
    type: bool
    version_added: 7.1.0
  fetch_attachments:
    description:
      - Boolean flag which indicates whether attached files are downloaded or not.
      - The download only happens if O(file_download_path) has been provided.
    required: false
    type: bool
    version_added: 7.0.0
  file_download_path:
    description: Indicate the file attachment download location.
    required: false
    type: path
    version_added: 7.0.0
  base_url:
    description: The base URL of the server, for example V(https://localhost/SecretServer).
    type: string
    env:
      - name: TSS_BASE_URL
    ini:
      - section: tss_lookup
        key: base_url
    required: true
  username:
    description: The username with which to request the OAuth2 Access Grant.
    type: string
    env:
      - name: TSS_USERNAME
    ini:
      - section: tss_lookup
        key: username
  password:
    description:
      - The password associated with the supplied username.
      - Required when O(token) is not provided.
    type: string
    env:
      - name: TSS_PASSWORD
    ini:
      - section: tss_lookup
        key: password
  domain:
    default: ""
    description:
      - The domain with which to request the OAuth2 Access Grant.
      - Optional when O(token) is not provided.
      - Requires C(python-tss-sdk) version 1.0.0 or greater.
    type: string
    env:
      - name: TSS_DOMAIN
    ini:
      - section: tss_lookup
        key: domain
    required: false
    version_added: 3.6.0
  token:
    description:
      - Existing token for Thycotic authorizer.
      - If provided, O(username) and O(password) are not needed.
      - Requires C(python-tss-sdk) version 1.0.0 or greater.
    type: string
    env:
      - name: TSS_TOKEN
    ini:
      - section: tss_lookup
        key: token
    version_added: 3.7.0
  api_path_uri:
    default: /api/v1
    description: The path to append to the base URL to form a valid REST API request.
    type: string
    env:
      - name: TSS_API_PATH_URI
    required: false
  token_path_uri:
    default: /oauth2/token
    description: The path to append to the base URL to form a valid OAuth2 Access Grant request.
    type: string
    env:
      - name: TSS_TOKEN_PATH_URI
    required: false
"""

RETURN = r"""
_list:
  description:
    - The JSON responses to C(GET /secrets/{id}) and C(GET /secrets/{path}).
    - See U(https://updates.thycotic.net/secretserver/restapiguide/TokenAuth/#operation--secrets--id--get).
  type: list
  elements: dict
"""

EXAMPLES = r"""
# Using Secret Server Authentication
- name: Lookup secret using Secret Server user credentials
  hosts: localhost
  vars:
    secret: >-
      {{
        lookup(
          'community.general.tss',
          102,
          base_url='https://secretserver.domain.com/SecretServer/',
          username='user.name',
          password='password'
        )
      }}
  tasks:
    - ansible.builtin.debug:
        msg: >
          the password is {{
            (secret['items']
              | items2dict(key_name='slug',
                           value_name='itemValue'))['password']
          }}

- name: Lookup secret with domain user
  hosts: localhost
  vars:
    secret: >-
      {{
        lookup(
          'community.general.tss',
          102,
          base_url='https://secretserver.domain.com/SecretServer/',
          username='user.name',
          password='password',
          domain='domain'
        )
      }}
  tasks:
    - ansible.builtin.debug:
        msg: >
          the password is {{
            (secret['items']
              | items2dict(key_name='slug',
                           value_name='itemValue'))['password']
          }}

- name: Lookup secret using Secret Server token
  hosts: localhost
  vars:
    secret_password: >-
      {{
        ((lookup(
          'community.general.tss',
          102,
          base_url='https://secretserver.domain.com/SecretServer/',
          token='thycotic_access_token',
        ) | from_json).get('items') | items2dict(key_name='slug', value_name='itemValue'))['password']
      }}
  tasks:
    - ansible.builtin.debug:
        msg: the password is {{ secret_password }}

# Private key stores into certificate file which is attached with secret.
# If fetch_attachments=True then private key file will be download on specified path
# and file content will display in debug message.
- name: Lookup secret and fetch attachments using Secret Server token
  hosts: localhost
  vars:
    secret: >-
      {{
        lookup(
          'community.general.tss',
          102,
          fetch_attachments=True,
          file_download_path='/home/certs',
          base_url='https://secretserver.domain.com/SecretServer/',
          token='thycotic_access_token'
        )
      }}
  tasks:
    - ansible.builtin.debug:
        msg: >
          the private key is {{
            (secret['items']
              | items2dict(key_name='slug',
                           value_name='itemValue'))['private-key']
          }}

# If fetch_secret_ids_from_folder=true then secret IDs are in a folder is fetched based on folder ID
- name: Lookup secret IDs by folder ID using Secret Server token
  hosts: localhost
  vars:
    secret: >-
      {{
        lookup(
          'community.general.tss',
          102,
          fetch_secret_ids_from_folder=true,
          base_url='https://secretserver.domain.com/SecretServer/',
          token='thycotic_access_token'
        )
      }}
  tasks:
    - ansible.builtin.debug:
        msg: >
          the secret id's are {{
              secret
          }}

# If secret ID is 0 and secret_path has value then secret is fetched by secret path
- name: Lookup secret by secret path using Secret Server user credentials
  hosts: localhost
  vars:
    secret: >-
      {{
        lookup(
          'community.general.tss',
          0,
          secret_path='\folderName\secretName'
          base_url='https://secretserver.domain.com/SecretServer/',
          username='user.name',
          password='password'
        )
      }}
  tasks:
    - ansible.builtin.debug:
        msg: >-
          the password is {{
            (secret['items']
              | items2dict(key_name='slug',
                           value_name='itemValue'))['password']
          }}

# Using Platform Authentication
- name: Lookup secret using Platform service user credentials
  hosts: localhost
  vars:
    secret: >-
      {{
        lookup(
          'community.general.tss',
          102,
          base_url='https://platform.delinea.app/',
          username='platform_service_username',
          password='platform_service_user_password'
        )
      }}
  tasks:
    - ansible.builtin.debug:
        msg: >
          the password is {{
            (secret['items']
              | items2dict(key_name='slug',
                           value_name='itemValue'))['password']
          }}

- name: Lookup secret using platform token
  hosts: localhost
  vars:
    secret_password: >-
      {{
        ((lookup(
          'community.general.tss',
          102,
          base_url='https://platform.delinea.app/',
          token='delinea_platform_access_token',
        ) | from_json).get('items') | items2dict(key_name='slug', value_name='itemValue'))['password']
      }}
  tasks:
    - ansible.builtin.debug:
        msg: the password is {{ secret_password }}
"""

import abc
import os

from ansible.errors import AnsibleError, AnsibleOptionsError
from ansible.plugins.lookup import LookupBase
from ansible.utils.display import Display

from ansible_collections.community.general.plugins.plugin_utils._lookup import check_for_wrong_terms

try:
    from delinea.secrets.server import (
        AccessTokenAuthorizer,
        DomainPasswordGrantAuthorizer,
        PasswordGrantAuthorizer,
        SecretServer,
        SecretServerError,
    )

    HAS_TSS_SDK = True
    HAS_DELINEA_SS_SDK = True
    HAS_TSS_AUTHORIZER = True
except ImportError:
    try:
        from thycotic.secrets.server import (
            AccessTokenAuthorizer,
            DomainPasswordGrantAuthorizer,
            PasswordGrantAuthorizer,
            SecretServer,
            SecretServerError,
        )

        HAS_TSS_SDK = True
        HAS_DELINEA_SS_SDK = False
        HAS_TSS_AUTHORIZER = True
    except ImportError:
        SecretServer = None
        SecretServerError = None
        HAS_TSS_SDK = False
        HAS_DELINEA_SS_SDK = False
        PasswordGrantAuthorizer = None
        DomainPasswordGrantAuthorizer = None
        AccessTokenAuthorizer = None
        HAS_TSS_AUTHORIZER = False


display = Display()


class TSSClient(metaclass=abc.ABCMeta):  # noqa: B024
    def __init__(self):
        self._client = None

    @staticmethod
    def from_params(**server_parameters):
        if HAS_TSS_AUTHORIZER:
            return TSSClientV1(**server_parameters)
        else:
            return TSSClientV0(**server_parameters)

    def get_secret(self, term, secret_path, fetch_file_attachments, file_download_path):
        display.debug(f"tss_lookup term: {term}")
        secret_id = self._term_to_secret_id(term)
        if secret_id == 0 and secret_path:
            fetch_secret_by_path = True
            display.vvv(f"Secret Server lookup of Secret with path {secret_path}")
        else:
            fetch_secret_by_path = False
            display.vvv(f"Secret Server lookup of Secret with ID {secret_id}")

        if fetch_file_attachments:
            if fetch_secret_by_path:
                obj = self._client.get_secret_by_path(secret_path, fetch_file_attachments)
            else:
                obj = self._client.get_secret(secret_id, fetch_file_attachments)
            for i in obj["items"]:
                if file_download_path and os.path.isdir(file_download_path):
                    if i["isFile"]:
                        try:
                            file_content = i["itemValue"].content
                            with open(os.path.join(file_download_path, f"{obj['id']}_{i['slug']}"), "wb") as f:
                                f.write(file_content)
                        except ValueError as e:
                            raise AnsibleOptionsError(f"Failed to download {i['slug']}") from e
                        except AttributeError:
                            display.warning(f"Could not read file content for {i['slug']}")
                        finally:
                            i["itemValue"] = "*** Not Valid For Display ***"
                else:
                    raise AnsibleOptionsError("File download path does not exist")
            return obj
        else:
            if fetch_secret_by_path:
                return self._client.get_secret_by_path(secret_path, False)
            else:
                return self._client.get_secret_json(secret_id)

    def get_secret_ids_by_folderid(self, term):
        display.debug(f"tss_lookup term: {term}")
        folder_id = self._term_to_folder_id(term)
        display.vvv(f"Secret Server lookup of Secret id's with Folder ID {folder_id}")

        return self._client.get_secret_ids_by_folderid(folder_id)

    @staticmethod
    def _term_to_secret_id(term):
        try:
            return int(term)
        except ValueError as e:
            raise AnsibleOptionsError("Secret ID must be an integer") from e

    @staticmethod
    def _term_to_folder_id(term):
        try:
            return int(term)
        except ValueError as e:
            raise AnsibleOptionsError("Folder ID must be an integer") from e


class TSSClientV0(TSSClient):
    def __init__(self, **server_parameters):
        super().__init__()

        if server_parameters.get("domain"):
            raise AnsibleError("The 'domain' option requires 'python-tss-sdk' version 1.0.0 or greater")

        self._client = SecretServer(
            server_parameters["base_url"],
            server_parameters["username"],
            server_parameters["password"],
            server_parameters["api_path_uri"],
            server_parameters["token_path_uri"],
        )


class TSSClientV1(TSSClient):
    def __init__(self, **server_parameters):
        super().__init__()

        authorizer = self._get_authorizer(**server_parameters)
        self._client = SecretServer(server_parameters["base_url"], authorizer, server_parameters["api_path_uri"])

    @staticmethod
    def _get_authorizer(**server_parameters):
        if server_parameters.get("token"):
            return AccessTokenAuthorizer(server_parameters["token"], server_parameters["base_url"])

        if server_parameters.get("domain"):
            return DomainPasswordGrantAuthorizer(
                server_parameters["base_url"],
                server_parameters["username"],
                server_parameters["domain"],
                server_parameters["password"],
                server_parameters["token_path_uri"],
            )

        return PasswordGrantAuthorizer(
            server_parameters["base_url"],
            server_parameters["username"],
            server_parameters["password"],
            server_parameters["token_path_uri"],
        )


class LookupModule(LookupBase):
    def run(self, terms, variables, **kwargs):
        if not HAS_TSS_SDK:
            raise AnsibleError("python-tss-sdk must be installed to use this plugin")

        self.set_options(var_options=variables, direct=kwargs)
        check_for_wrong_terms(self, direct=kwargs)

        tss = TSSClient.from_params(
            base_url=self.get_option("base_url"),
            username=self.get_option("username"),
            password=self.get_option("password"),
            domain=self.get_option("domain"),
            token=self.get_option("token"),
            api_path_uri=self.get_option("api_path_uri"),
            token_path_uri=self.get_option("token_path_uri"),
        )

        try:
            if self.get_option("fetch_secret_ids_from_folder"):
                if HAS_DELINEA_SS_SDK:
                    return [tss.get_secret_ids_by_folderid(term) for term in terms]
                else:
                    raise AnsibleError("latest python-tss-sdk must be installed to use this plugin")
            else:
                return [
                    tss.get_secret(
                        term,
                        self.get_option("secret_path"),
                        self.get_option("fetch_attachments"),
                        self.get_option("file_download_path"),
                    )
                    for term in terms
                ]
        except SecretServerError as error:
            raise AnsibleError(f"Secret Server lookup failure: {error.message}") from error



# -*- coding: utf-8 -*-
# Copyright 2025 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function


__metaclass__ = type

"""
The L2_interfaces parser templates file. This contains
a list of parser definitions and associated functions that
facilitates both facts gathering and native command generation for
the given network resource.
"""

import re

from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.rm_base.network_template import (
    NetworkTemplate,
)


class L2_interfacesTemplate(NetworkTemplate):
    def __init__(self, lines=None, module=None):
        super(L2_interfacesTemplate, self).__init__(lines=lines, tmplt=self, module=module)

    # fmt: off
    PARSERS = [
        {
            "name": "name",
            "getval": re.compile(
                r"""
                ^interface\s(?P<name>\S+)
                $""", re.VERBOSE,
            ),
            "setval": 'interface {{ name }}',
            "result": {
                '{{ name }}': {
                    'name': '{{ name }}',
                },
            },
            "shared": True,
        },
        {
            "name": "mode",
            "getval": re.compile(
                r"""
                \s+switchport\smode
                \s+(?P<mode>access|trunk|dot1q-tunnel|fex-fabric|fabricpath)
                $""", re.VERBOSE,
            ),
            "setval": "switchport mode {{ mode }}",
            "result": {
                '{{ name }}': {
                    'mode': "{{ mode }}",
                },
            },
        },
        {
            "name": "access.vlan",
            "getval": re.compile(
                r"""
                \s+switchport\saccess\svlan
                \s+(?P<vlan>\d+)
                $""", re.VERBOSE,
            ),
            "setval": "switchport access vlan {{ access.vlan }}",
            "result": {
                '{{ name }}': {
                    'access': {
                        'vlan': "{{ vlan }}",
                    },
                },
            },
        },
        {
            "name": "trunk.native_vlan",
            "getval": re.compile(
                r"""
                \s+switchport\strunk\snative\svlan
                \s+(?P<vlan>\d+)
                $""", re.VERBOSE,
            ),
            "setval": "switchport trunk native vlan {{ trunk.native_vlan|string }}",
            "result": {
                '{{ name }}': {
                    'trunk': {
                        'native_vlan': "{{ vlan }}",
                    },
                },
            },
        },
        {
            "name": "trunk.allowed_vlans_none",
            "getval": re.compile(
                r"""
                \s+switchport\strunk\sallowed\svlan
                \s+(?P<isnone>none)
                $""", re.VERBOSE,
            ),
            "setval": "switchport trunk allowed vlan",
            "result": {
                '{{ name }}': {
                    'trunk': {
                        'allowed_vlans_none': "{{ not not isnone }}",
                    },
                },
            },
        },
        {
            "name": "trunk.allowed_vlans",
            "getval": re.compile(
                r"""
                \s+switchport\strunk\sallowed\svlan
                \s+(?!none$)(?P<allowed_vlans>.+)
                $""", re.VERBOSE,
            ),
            "setval": "",
            "result": {
                '{{ name }}': {
                    'trunk': {
                        'allowed_vlans': "'{{ allowed_vlans }}'",
                    },
                },
            },
        },
        {
            "name": "beacon",
            "getval": re.compile(
                r"""
                \s+(?P<beacon>beacon)
                $""", re.VERBOSE,
            ),
            "setval": "beacon",
            "result": {
                '{{ name }}': {
                    'beacon': "{{ True if beacon }}",
                },
            },
        },
        {
            "name": "link_flap.error_disable",
            "getval": re.compile(
                r"""
                \s+link-flap\serror-disable
                (\s+count\s(?P<count>\d+))?
                (\s+interval\s(?P<interval>\d+))?
                $""", re.VERBOSE,
            ),
            "setval": "link-flap error-disable"
            "{{ ' count ' + count|string if count is defined else '' }}"
            "{{ ' interval ' + interval|string if interval is defined else '' }}",
            "result": {
                '{{ name }}': {
                    'link_flap': {
                        'error_disable': {
                            'count': "{{ count }}",
                            'interval': "{{ interval }}",
                        },
                    },
                },
            },
        },
        {
            "name": "cdp_enable",
            "getval": re.compile(
                r"""
                \s+cdp\senable
                $""", re.VERBOSE,
            ),
            "setval": "cdp enable",
            "result": {
                '{{ name }}': {
                    'cdp_enable': True,
                },
            },
        },
        {
            "name": "no_cdp_enable",
            "getval": re.compile(
                r"""
                \s+no\scdp\senable
                $""", re.VERBOSE,
            ),
            "setval": "no cdp enable",
            "result": {
                '{{ name }}': {
                    'cdp_enable': False,
                },
            },
        },
    ]
    # fmt: on



#!/usr/bin/python
# -*- coding: utf-8 -*-

# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}

DOCUMENTATION = r'''
---
module: intersight_fibre_channel_qos_policy
short_description: Manage Fibre Channel QoS Policies for Cisco Intersight
description:
  - Create, update, and delete Fibre Channel QoS Policies on Cisco Intersight.
  - Fibre Channel QoS policies configure Quality of Service settings for Fibre Channel virtual interfaces.
  - These policies control bandwidth rate limits, maximum data field size, class of service, and burst traffic.
  - For more information see L(Cisco Intersight,https://intersight.com/apidocs/vnic/FcQosPolicies/get/).
extends_documentation_fragment: intersight
options:
  state:
    description:
      - If C(present), will verify the resource is present and will create if needed.
      - If C(absent), will verify the resource is absent and will delete if needed.
    type: str
    choices: [present, absent]
    default: present
  organization:
    description:
      - The name of the Organization this resource is assigned to.
      - Policies created within a Custom Organization are applicable only to devices in the same Organization.
      - Use 'default' for the default organization.
    type: str
    default: default
  name:
    description:
      - The name assigned to the Fibre Channel QoS Policy.
      - Must be unique within the organization.
      - The name must be between 1 and 64 alphanumeric characters, allowing special characters :-_.
    type: str
    required: true
  description:
    description:
      - The user-defined description for the Fibre Channel QoS Policy.
      - Description can contain letters(a-z, A-Z), numbers(0-9), hyphen(-), period(.), colon(:), or an underscore(_).
    type: str
    aliases: [descr]
  tags:
    description:
      - List of tags in Key:<user-defined key> Value:<user-defined value> format.
    type: list
    elements: dict
  rate_limit:
    description:
      - The value in Mbps to use for limiting the data rate on the virtual interface.
      - A value of 0 means no rate limiting.
      - Valid range is 0 to 100000 Mbps.
    type: int
    default: 0
  max_data_field_size:
    description:
      - The maximum size of the Fibre Channel frame payload bytes that the virtual interface supports.
      - Valid range is 256 to 2112 bytes.
    type: int
    default: 2112
  cos:
    description:
      - Class of Service to be associated to the traffic on the virtual interface.
      - Valid range is 0 to 6.
    type: int
    default: 3
  burst:
    description:
      - The burst traffic, in bytes, allowed on the vHBA.
      - Valid range is 1 to 1000000 bytes.
    type: int
    default: 10240
author:
  - Ron Gershburg (@rgershbu)
'''

EXAMPLES = r'''
- name: Create Fibre Channel QoS Policy with default settings
  cisco.intersight.intersight_fibre_channel_qos_policy:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
    organization: "default"
    name: "fc-qos-default"
    description: "Fibre Channel QoS policy with default values"
    state: present

- name: Create Fibre Channel QoS Policy with custom settings
  cisco.intersight.intersight_fibre_channel_qos_policy:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
    organization: "default"
    name: "fc-qos-custom"
    description: "Fibre Channel QoS policy with custom values"
    rate_limit: 10000
    max_data_field_size: 2048
    cos: 5
    burst: 20480
    tags:
      - Key: Environment
        Value: Production
    state: present

- name: Create Fibre Channel QoS Policy with maximum rate limit
  cisco.intersight.intersight_fibre_channel_qos_policy:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
    organization: "Engineering"
    name: "fc-qos-high-bandwidth"
    description: "Fibre Channel QoS policy with maximum bandwidth"
    rate_limit: 100000
    max_data_field_size: 2112
    cos: 6
    burst: 1000000
    state: present

- name: Create Fibre Channel QoS Policy with minimal frame size
  cisco.intersight.intersight_fibre_channel_qos_policy:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
    organization: "default"
    name: "fc-qos-minimal"
    description: "Fibre Channel QoS policy with minimal frame size"
    rate_limit: 0
    max_data_field_size: 256
    cos: 0
    burst: 1
    state: present

- name: Update Fibre Channel QoS Policy description
  cisco.intersight.intersight_fibre_channel_qos_policy:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
    organization: "default"
    name: "fc-qos-default"
    description: "Updated Fibre Channel QoS policy description"
    state: present

- name: Delete Fibre Channel QoS Policy
  cisco.intersight.intersight_fibre_channel_qos_policy:
    api_private_key: "{{ api_private_key }}"
    api_key_id: "{{ api_key_id }}"
    organization: "default"
    name: "fc-qos-default"
    state: absent
'''

RETURN = r'''
api_response:
  description: The API response output returned by the specified resource.
  returned: always
  type: dict
  sample:
    "api_response": {
        "Name": "fc-qos-custom",
        "ObjectType": "vnic.FcQosPolicy",
        "RateLimit": 10000,
        "MaxDataFieldSize": 2048,
        "Cos": 5,
        "Burst": 20480,
        "Moid": "1234567890abcdef12345678",
        "Description": "Fibre Channel QoS policy with custom values",
        "Organization": {
            "Moid": "abcdef1234567890abcdef12",
            "ObjectType": "organization.Organization"
        },
        "Tags": [
            {
                "Key": "Environment",
                "Value": "Production"
            }
        ]
    }
'''


from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.intersight.plugins.module_utils.intersight import IntersightModule, intersight_argument_spec


def validate_parameters(module):
    """
    Validate module parameters for Fibre Channel QoS policy configuration.
    """
    if module.params['state'] != 'present':
        return
    # Validate rate_limit range (0-100000)
    rate_limit = module.params.get('rate_limit')
    if rate_limit is not None and (rate_limit < 0 or rate_limit > 100000):
        module.fail_json(msg="Parameter 'rate_limit' must be between 0 and 100000 Mbps")
    # Validate max_data_field_size range (256-2112)
    max_data_field_size = module.params.get('max_data_field_size')
    if max_data_field_size is not None and (max_data_field_size < 256 or max_data_field_size > 2112):
        module.fail_json(msg="Parameter 'max_data_field_size' must be between 256 and 2112 bytes")
    # Validate cos range (0-6)
    cos = module.params.get('cos')
    if cos is not None and (cos < 0 or cos > 6):
        module.fail_json(msg="Parameter 'cos' must be between 0 and 6")
    # Validate burst range (1-1000000)
    burst = module.params.get('burst')
    if burst is not None and (burst < 1 or burst > 1000000):
        module.fail_json(msg="Parameter 'burst' must be between 1 and 1000000 bytes")


def main():
    argument_spec = intersight_argument_spec.copy()
    argument_spec.update(
        state=dict(type='str', choices=['present', 'absent'], default='present'),
        organization=dict(type='str', default='default'),
        name=dict(type='str', required=True),
        description=dict(type='str', aliases=['descr']),
        tags=dict(type='list', elements='dict'),
        rate_limit=dict(type='int', default=0),
        max_data_field_size=dict(type='int', default=2112),
        cos=dict(type='int', default=3),
        burst=dict(type='int', default=10240),
    )
    module = AnsibleModule(
        argument_spec,
        supports_check_mode=True,
    )
    intersight = IntersightModule(module)
    intersight.result['api_response'] = {}
    intersight.result['trace_id'] = ''
    # Validate module parameters
    validate_parameters(module)
    # Resource path used to configure policy
    resource_path = '/vnic/FcQosPolicies'
    # Define API body used in compares or create
    intersight.api_body = {
        'Organization': {
            'Name': module.params['organization'],
        },
        'Name': module.params['name'],
    }
    if module.params['state'] == 'present':
        intersight.api_body['RateLimit'] = module.params['rate_limit']
        intersight.api_body['MaxDataFieldSize'] = module.params['max_data_field_size']
        intersight.api_body['Cos'] = module.params['cos']
        intersight.api_body['Burst'] = module.params['burst']
        intersight.set_tags_and_description()
    intersight.configure_policy_or_profile(resource_path=resource_path)
    module.exit_json(**intersight.result)


if __name__ == '__main__':
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2019 Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

# Author:
#   - Matthew Davis <Matthew.Davis.2@team.telstra.com>
#     on behalf of Telstra Corporation Limited

DOCUMENTATION = r"""
---
module: acm_certificate
short_description: Upload and delete certificates in the AWS Certificate Manager service
version_added: 1.0.0
description:
  - >
    Import and delete certificates in Amazon Web Service's Certificate
    Manager (AWS ACM).
  - >
    This module does not currently interact with AWS-provided certificates.
    It currently only manages certificates provided to AWS by the user.
  - The ACM API allows users to upload multiple certificates for the same domain
    name, and even multiple identical certificates. This module attempts to
    restrict such freedoms, to be idempotent, as per the Ansible philosophy.
    It does this through applying AWS resource "Name" tags to ACM certificates.
  - >
    When I(state=present),
    if there is one certificate in ACM
    with a C(Name) tag equal to the I(name_tag) parameter,
    and an identical body and chain,
    this task will succeed without effect.
  - >
    When I(state=present),
    if there is one certificate in ACM
    a I(Name) tag equal to the I(name_tag) parameter,
    and a different body,
    this task will overwrite that certificate.
  - >
    When I(state=present),
    if there are multiple certificates in ACM
    with a I(Name) tag equal to the I(name_tag) parameter,
    this task will fail.
  - >
    When I(state=absent) and I(certificate_arn) is defined,
    this module will delete the ACM resource with that ARN if it exists in this
    region, and succeed without effect if it doesn't exist.
  - >
    When I(state=absent) and I(domain_name) is defined, this module will delete
    all ACM resources in this AWS region with a corresponding domain name.
    If there are none, it will succeed without effect.
  - >
    When I(state=absent) and I(certificate_arn) is not defined,
    and I(domain_name) is not defined, this module will delete all ACM resources
    in this AWS region with a corresponding I(Name) tag.
    If there are none, it will succeed without effect.
  - >
    Note that this may not work properly with keys of size 4096 bits, due to a
    limitation of the ACM API.
  - Prior to release 5.0.0 this module was called C(community.aws.aws_acm).
    The usage did not change.
options:
  certificate:
    description:
      - The body of the PEM encoded public certificate.
      - Required when I(state) is not C(absent) and the certificate does not exist.
      - >
        If your certificate is in a file,
        use C(lookup('file', 'path/to/cert.pem')).
    type: str
  certificate_arn:
    description:
      - The ARN of a certificate in ACM to modify or delete.
      - >
        If I(state=present), the certificate with the specified ARN can be updated.
        For example, this can be used to add/remove tags to an existing certificate.
      - >
        If I(state=absent), you must provide one of
        I(certificate_arn), I(domain_name) or I(name_tag).
      - >
        If I(state=absent) and no resource exists with this ARN in this region,
        the task will succeed with no effect.
      - >
        If I(state=absent) and the corresponding resource exists in a different
        region, this task may report success without deleting that resource.
    type: str
    aliases: [arn]
  certificate_chain:
    description:
      - The body of the PEM encoded chain for your certificate.
      - >
        If your certificate chain is in a file,
        use C(lookup('file', 'path/to/chain.pem')).
      - Ignored when I(state=absent)
    type: str
  domain_name:
    description:
      - The domain name of the certificate.
      - >
        If I(state=absent) and I(domain_name) is specified,
        this task will delete all ACM certificates with this domain.
      - >
        Exactly one of I(domain_name), I(name_tag) and I(certificate_arn)
        must be provided.
      - >
        If I(state=present) this must not be specified.
        (Since the domain name is encoded within the public certificate's body.)
    type: str
    aliases: [domain]
  name_tag:
    description:
      - >
        The unique identifier for tagging resources using AWS tags,
        with key I(Name).
      - This can be any set of characters accepted by AWS for tag values.
      - >
        This is to ensure Ansible can treat certificates idempotently,
        even though the ACM API allows duplicate certificates.
      - If I(state=preset), this must be specified.
      - >
        If I(state=absent) and I(name_tag) is specified,
        this task will delete all ACM certificates with this Name tag.
      - >
        If I(state=absent), you must provide exactly one of
        I(certificate_arn), I(domain_name) or I(name_tag).
      - >
        If both I(name_tag) and the 'Name' tag in I(tags) are set,
        the values must be the same.
      - >
        If the 'Name' tag in I(tags) is not set and I(name_tag) is set,
        the I(name_tag) value is copied to I(tags).
    type: str
    aliases: [name]
  private_key:
    description:
      - The body of the PEM encoded private key.
      - Required when I(state=present) and the certificate does not exist.
      - Ignored when I(state=absent).
      - >
        If your private key is in a file,
        use C(lookup('file', 'path/to/key.pem')).
    type: str
  state:
    description:
      - >
        If I(state=present), the specified public certificate and private key
        will be uploaded, with I(Name) tag equal to I(name_tag).
      - >
        If I(state=absent), any certificates in this region
        with a corresponding I(domain_name), I(name_tag) or I(certificate_arn)
        will be deleted.
    choices: [present, absent]
    default: present
    type: str

notes:
  - Support for I(tags) and I(purge_tags) was added in release 3.2.0
author:
  - Matthew Davis (@matt-telstra) on behalf of Telstra Corporation Limited
extends_documentation_fragment:
  - amazon.aws.common.modules
  - amazon.aws.region.modules
  - amazon.aws.tags
  - amazon.aws.boto3
"""

EXAMPLES = r"""

- name: upload a self-signed certificate
  community.aws.acm_certificate:
    certificate: "{{ lookup('file', 'cert.pem' ) }}"
    privateKey: "{{ lookup('file', 'key.pem' ) }}"
    name_tag: my_cert # to be applied through an AWS tag as  "Name":"my_cert"
    region: ap-southeast-2 # AWS region

- name: create/update a certificate with a chain
  community.aws.acm_certificate:
    certificate: "{{ lookup('file', 'cert.pem' ) }}"
    private_key: "{{ lookup('file', 'key.pem' ) }}"
    name_tag: my_cert
    certificate_chain: "{{ lookup('file', 'chain.pem' ) }}"
    state: present
    region: ap-southeast-2
  register: cert_create

- name: print ARN of cert we just created
  ansible.builtin.debug:
    var: cert_create.certificate.arn

- name: delete the cert we just created
  community.aws.acm_certificate:
    name_tag: my_cert
    state: absent
    region: ap-southeast-2

- name: delete a certificate with a particular ARN
  community.aws.acm_certificate:
    certificate_arn: "arn:aws:acm:ap-southeast-2:123456789012:certificate/01234567-abcd-abcd-abcd-012345678901"
    state: absent
    region: ap-southeast-2

- name: delete all certificates with a particular domain name
  community.aws.acm_certificate:
    domain_name: acm.ansible.com
    state: absent
    region: ap-southeast-2

- name: add tags to an existing certificate with a particular ARN
  community.aws.acm_certificate:
    certificate_arn: "arn:aws:acm:ap-southeast-2:123456789012:certificate/01234567-abcd-abcd-abcd-012345678901"
    tags:
      Name: my_certificate
      Application: search
      Environment: development
    purge_tags: true
"""

RETURN = r"""
certificate:
  description: Information about the certificate which was uploaded
  type: complex
  returned: when I(state=present)
  contains:
    arn:
      description: The ARN of the certificate in ACM
      type: str
      returned: when I(state=present) and not in check mode
      sample: "arn:aws:acm:ap-southeast-2:123456789012:certificate/01234567-abcd-abcd-abcd-012345678901"
    domain_name:
      description: The domain name encoded within the public certificate
      type: str
      returned: when I(state=present)
      sample: acm.ansible.com
arns:
  description: A list of the ARNs of the certificates in ACM which were deleted
  type: list
  elements: str
  returned: when I(state=absent)
  sample:
   - "arn:aws:acm:ap-southeast-2:123456789012:certificate/01234567-abcd-abcd-abcd-012345678901"
"""


import base64
import re  # regex library
from copy import deepcopy

try:
    import botocore
except ImportError:
    pass  # handled by AnsibleAWSModule

from ansible.module_utils._text import to_text

from ansible_collections.amazon.aws.plugins.module_utils.acm import ACMServiceManager
from ansible_collections.amazon.aws.plugins.module_utils.tagging import ansible_dict_to_boto3_tag_list
from ansible_collections.amazon.aws.plugins.module_utils.tagging import boto3_tag_list_to_ansible_dict
from ansible_collections.amazon.aws.plugins.module_utils.tagging import compare_aws_tags

from ansible_collections.community.aws.plugins.module_utils.modules import AnsibleCommunityAWSModule as AnsibleAWSModule


def ensure_tags(client, module, resource_arn, existing_tags, tags, purge_tags):
    if tags is None:
        return (False, existing_tags)

    tags_to_add, tags_to_remove = compare_aws_tags(existing_tags, tags, purge_tags)
    changed = bool(tags_to_add or tags_to_remove)
    if tags_to_add and not module.check_mode:
        try:
            client.add_tags_to_certificate(
                CertificateArn=resource_arn,
                Tags=ansible_dict_to_boto3_tag_list(tags_to_add),
            )
        except (
            botocore.exceptions.ClientError,
            botocore.exceptions.BotoCoreError,
        ) as e:
            module.fail_json_aws(e, f"Couldn't add tags to certificate {resource_arn}")
    if tags_to_remove and not module.check_mode:
        # remove_tags_from_certificate wants a list of key, value pairs, not a list of keys.
        tags_list = [{"Key": key, "Value": existing_tags.get(key)} for key in tags_to_remove]
        try:
            client.remove_tags_from_certificate(
                CertificateArn=resource_arn,
                Tags=tags_list,
            )
        except (
            botocore.exceptions.ClientError,
            botocore.exceptions.BotoCoreError,
        ) as e:
            module.fail_json_aws(e, f"Couldn't remove tags from certificate {resource_arn}")
    new_tags = deepcopy(existing_tags)
    for key, value in tags_to_add.items():
        new_tags[key] = value
    for key in tags_to_remove:
        new_tags.pop(key, None)
    return (changed, new_tags)


# Takes in two text arguments
# Each a PEM encoded certificate
# Or a chain of PEM encoded certificates
# May include some lines between each chain in the cert, e.g. "Subject: ..."
# Returns True iff the chains/certs are functionally identical (including chain order)
def chain_compare(module, a, b):
    chain_a_pem = pem_chain_split(module, a)
    chain_b_pem = pem_chain_split(module, b)

    if len(chain_a_pem) != len(chain_b_pem):
        return False

    # Chain length is the same
    for ca, cb in zip(chain_a_pem, chain_b_pem):
        der_a = PEM_body_to_DER(module, ca)
        der_b = PEM_body_to_DER(module, cb)
        if der_a != der_b:
            return False

    return True


# Takes in PEM encoded data with no headers
# returns equivilent DER as byte array
def PEM_body_to_DER(module, pem):
    try:
        der = base64.b64decode(to_text(pem))
    except (ValueError, TypeError) as e:
        module.fail_json_aws(e, msg="Unable to decode certificate chain")
    return der


# Store this globally to avoid repeated recompilation
pem_chain_split_regex = re.compile(
    r"------?BEGIN [A-Z0-9. ]*CERTIFICATE------?([a-zA-Z0-9\+\/=\s]+)------?END [A-Z0-9. ]*CERTIFICATE------?"
)


# Use regex to split up a chain or single cert into an array of base64 encoded data
# Using "-----BEGIN CERTIFICATE-----" and "----END CERTIFICATE----"
# Noting that some chains have non-pem data in between each cert
# This function returns only what's between the headers, excluding the headers
def pem_chain_split(module, pem):
    pem_arr = re.findall(pem_chain_split_regex, to_text(pem))

    if len(pem_arr) == 0:
        # This happens if the regex doesn't match at all
        module.fail_json(msg="Unable to split certificate chain. Possibly zero-length chain?")

    return pem_arr


def update_imported_certificate(client, module, acm, old_cert, desired_tags):
    """
    Update the existing certificate that was previously imported in ACM.
    """
    module.debug("Existing certificate found in ACM")
    if ("tags" not in old_cert) or ("Name" not in old_cert["tags"]):
        # shouldn't happen
        module.fail_json(msg="Internal error, unsure which certificate to update", certificate=old_cert)
    if module.params.get("name_tag") is not None and (old_cert["tags"]["Name"] != module.params.get("name_tag")):
        # This could happen if the user identified the certificate using 'certificate_arn' or 'domain_name',
        # and the 'Name' tag in the AWS API does not match the ansible 'name_tag'.
        module.fail_json(msg="Internal error, Name tag does not match", certificate=old_cert)
    if "certificate" not in old_cert:
        # shouldn't happen
        module.fail_json(msg="Internal error, unsure what the existing cert in ACM is", certificate=old_cert)

    cert_arn = None
    # Are the existing certificate in ACM and the local certificate the same?
    same = True
    if module.params.get("certificate") is not None:
        same &= chain_compare(module, old_cert["certificate"], module.params["certificate"])
        if module.params["certificate_chain"]:
            # Need to test this
            # not sure if Amazon appends the cert itself to the chain when self-signed
            same &= chain_compare(module, old_cert["certificate_chain"], module.params["certificate_chain"])
        else:
            # When there is no chain with a cert
            # it seems Amazon returns the cert itself as the chain
            same &= chain_compare(module, old_cert["certificate_chain"], module.params["certificate"])

    if same:
        module.debug("Existing certificate in ACM is the same")
        cert_arn = old_cert["certificate_arn"]
        changed = False
    else:
        absent_args = ["certificate", "name_tag", "private_key"]
        if sum([(module.params[a] is not None) for a in absent_args]) < 3:
            module.fail_json(
                msg="When importing a certificate, all of 'name_tag', 'certificate' and 'private_key' must be specified"
            )
        module.debug("Existing certificate in ACM is different, overwriting")
        changed = True
        if module.check_mode:
            cert_arn = old_cert["certificate_arn"]
            # note: returned domain will be the domain of the previous cert
        else:
            # update cert in ACM
            cert_arn = acm.import_certificate(
                client,
                module,
                certificate=module.params["certificate"],
                private_key=module.params["private_key"],
                certificate_chain=module.params["certificate_chain"],
                arn=old_cert["certificate_arn"],
                tags=desired_tags,
            )
    return (changed, cert_arn)


def import_certificate(client, module, acm, desired_tags):
    """
    Import a certificate to ACM.
    """
    # Validate argument requirements
    absent_args = ["certificate", "name_tag", "private_key"]
    cert_arn = None
    if sum([(module.params[a] is not None) for a in absent_args]) < 3:
        module.fail_json(
            msg="When importing a new certificate, all of 'name_tag', 'certificate' and 'private_key' must be specified"
        )
    module.debug("No certificate in ACM. Creating new one.")
    changed = True
    if module.check_mode:
        domain = "example.com"
        module.exit_json(certificate=dict(domain_name=domain), changed=True)
    else:
        cert_arn = acm.import_certificate(
            client,
            module,
            certificate=module.params["certificate"],
            private_key=module.params["private_key"],
            certificate_chain=module.params["certificate_chain"],
            tags=desired_tags,
        )
    return (changed, cert_arn)


def ensure_certificates_present(client, module, acm, certificates, desired_tags, filter_tags):
    cert_arn = None
    changed = False
    if len(certificates) > 1:
        msg = f"More than one certificate with Name={module.params['name_tag']} exists in ACM in this region"
        module.fail_json(msg=msg, certificates=certificates)
    elif len(certificates) == 1:
        # Update existing certificate that was previously imported to ACM.
        (changed, cert_arn) = update_imported_certificate(client, module, acm, certificates[0], desired_tags)
    else:  # len(certificates) == 0
        # Import new certificate to ACM.
        (changed, cert_arn) = import_certificate(client, module, acm, desired_tags)

    # Add/remove tags to/from certificate
    try:
        existing_tags = boto3_tag_list_to_ansible_dict(
            client.list_tags_for_certificate(CertificateArn=cert_arn)["Tags"]
        )
    except (botocore.exceptions.ClientError, botocore.exceptions.BotoCoreError) as e:
        module.fail_json_aws(e, "Couldn't get tags for certificate")

    purge_tags = module.params.get("purge_tags")
    (c, new_tags) = ensure_tags(client, module, cert_arn, existing_tags, desired_tags, purge_tags)
    changed |= c
    domain = acm.get_domain_of_cert(client=client, module=module, arn=cert_arn)
    module.exit_json(certificate=dict(domain_name=domain, arn=cert_arn, tags=new_tags), changed=changed)


def ensure_certificates_absent(client, module, acm, certificates):
    for cert in certificates:
        if not module.check_mode:
            acm.delete_certificate(client, module, cert["certificate_arn"])
    module.exit_json(arns=[cert["certificate_arn"] for cert in certificates], changed=(len(certificates) > 0))


def main():
    argument_spec = dict(
        certificate=dict(),
        certificate_arn=dict(aliases=["arn"]),
        certificate_chain=dict(),
        domain_name=dict(aliases=["domain"]),
        name_tag=dict(aliases=["name"]),
        private_key=dict(no_log=True),
        tags=dict(type="dict", aliases=["resource_tags"]),
        purge_tags=dict(type="bool", default=True),
        state=dict(default="present", choices=["present", "absent"]),
    )
    module = AnsibleAWSModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
    )
    acm = ACMServiceManager(module)

    # Check argument requirements
    if module.params["state"] == "present":
        # at least one of these should be specified.
        absent_args = ["certificate_arn", "domain_name", "name_tag"]
        if sum([(module.params[a] is not None) for a in absent_args]) < 1:
            for a in absent_args:
                module.debug(f"{a} is {module.params[a]}")
            module.fail_json(
                msg="If 'state' is specified as 'present' then at least one of 'name_tag', 'certificate_arn' or 'domain_name' must be specified"
            )
    else:  # absent
        # exactly one of these should be specified
        absent_args = ["certificate_arn", "domain_name", "name_tag"]
        if sum([(module.params[a] is not None) for a in absent_args]) != 1:
            for a in absent_args:
                module.debug(f"{a} is {module.params[a]}")
            module.fail_json(
                msg="If 'state' is specified as 'absent' then exactly one of 'name_tag', 'certificate_arn' or 'domain_name' must be specified"
            )

    filter_tags = None
    desired_tags = None
    if module.params.get("tags") is not None:
        desired_tags = module.params["tags"]
    else:
        # Because we're setting the Name tag, we need to explicitly not purge when tags isn't passed
        module.params["purge_tags"] = False
    if module.params.get("name_tag") is not None:
        # The module was originally implemented to filter certificates based on the 'Name' tag.
        # Other tags are not used to filter certificates.
        # It would make sense to replace the existing name_tag, domain, certificate_arn attributes
        # with a 'filter' attribute, but that would break backwards-compatibility.
        filter_tags = dict(Name=module.params["name_tag"])
        if desired_tags is not None:
            if "Name" in desired_tags:
                if desired_tags["Name"] != module.params["name_tag"]:
                    module.fail_json(msg="Value of 'name_tag' conflicts with value of 'tags.Name'")
            else:
                desired_tags["Name"] = module.params["name_tag"]
        else:
            desired_tags = deepcopy(filter_tags)

    client = module.client("acm")

    # fetch the list of certificates currently in ACM
    certificates = acm.get_certificates(
        client=client,
        module=module,
        domain_name=module.params["domain_name"],
        arn=module.params["certificate_arn"],
        only_tags=filter_tags,
    )

    module.debug(f"Found {len(certificates)} corresponding certificates in ACM")
    if module.params["state"] == "present":
        ensure_certificates_present(client, module, acm, certificates, desired_tags, filter_tags)

    else:  # state == absent
        ensure_certificates_absent(client, module, acm, certificates)


if __name__ == "__main__":
    # tests()
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2024, Aleksandr Vagachev (@aleksvagachev) <aleksvagachev@yandex.ru>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import (absolute_import, division, print_function)
__metaclass__ = type

DOCUMENTATION = r'''
---
module: clickhouse_user

short_description: Creates or removes a ClickHouse user using the clickhouse-driver Client interface

description:
  - Creates or removes a ClickHouse user using the
    L(clickhouse-driver,https://clickhouse-driver.readthedocs.io/en/latest) Client interface.
  - The module can only create and delete users, without any additional parameters.
    New features will be added in the future.

attributes:
  check_mode:
    description: Supports check_mode.
    support: full
  idempotent:
    support: partial
    description:
      - O(settings) can break idempotency when the deprecated C(list) format is used. Use the dictionary format instead.
      - O(update_password) turns off idempotency because ClickHouse doesn't expose password hashes and the update will always be executed.

author:
  - Aleksandr Vagachev (@aleksvagachev)
  - Andrew Klychkov (@Andersson007)
  - Rafal Kozlowski (@rkozlo)

extends_documentation_fragment:
  - community.clickhouse.client_inst_opts

version_added: '0.4.0'

options:
  state:
    description:
      - User state.
      - If C(present), will create the user if not exists.
      - If C(absent), will drop the user if exists.
    type: str
    choices: ['present', 'absent']
    default: 'present'
  name:
    description:
      - User name to add or remove.
    type: str
    required: true
  password:
    description:
      - This option is deprecated and will be removed in 3.0.0. Use O(authentication) instead.
      - Set the user's password.
      - Password can be passed unhashed or hashed.
      - Cannot be used together with O(authentication).
    type: str
  type_password:
    description:
      - This option is deprecated and will be removed in 3.0.0. Use O(authentication) instead.
      - The type of password being transmitted(plaintext_password, sha256_password, sha256_hash...).
      - For more details, see U(https://clickhouse.com/docs/en/sql-reference/statements/create/user).
      - Cannot be used together with O(authentication).
    type: str
    default: sha256_password
  authentication:
    version_added: '2.1.0'
    description:
      - Set the user authentication.
      - Allows using LDAP, different hashing password.
      - Cannot be used together with legacy O(password) and O(type_password).
    type: dict
    suboptions:
      type:
        description: Authentication method to use.
        default: sha256_password
        type: str
        choices:
          - not_identified
          - no_password
          - plaintext_password
          - sha256_password
          - sha256_hash
          - double_sha1_password
          - double_sha1_hash
          - bcrypt_password
          - bcrypt_hash
          - ldap
      server:
        description:
          - Name of the server (required when I(type=ldap)).
        type: str
      password:
        description:
          - Password or hash.
        type: str
  cluster:
    description:
      - Run the command on all cluster hosts.
      - If the cluster is not configured, the command will crash with an error.
    type: str
  update_password:
    description:
      - If C(on_create), will set the password only for newly created users.
        If the user already exists, a C(password) value will be ignored.
      - If C(always), will always update the password.
        This option is not idempotent and will update the password even
        if it is the same in the database. If in future ClickHouse will allow
        to retrieve password hashes and other necessary details, this behavior
        will be changed.
    type: str
    choices: [always, on_create]
    default: on_create
  settings:
    description:
      - Settings with their constraints applied by default at user login.
      - Can pass either C(list)(obsolete) or dictionary.
      - You can also specify the profile from which the settings will be inherited C(list).
      - When specified for an existing user, settings will only be updated if they differ from current settings.
      - The module fetches current settings from C(system.settings_profile_elements) for comparison.
      - If no O(profiles) or O(settings) defined setting will not remove anything. Set to {} to purge it.
    type: raw
    version_added: '0.5.0'
  roles:
    description:
      - Grants specified roles to the user.
      - To append or remove roles, use the O(roles_mode) argument.
      - To revoke all roles, pass an empty list (C([])) and O(default_roles_mode=listed_only).
    type: list
    elements: str
    version_added: '0.6.0'
  default_roles:
    description:
      - Sets specified roles as default for the user.
      - The roles must be explicitly granted to the user whether manually
        before using this argument or by using the O(roles)
        argument in the same task.
      - To append or remove roles, use the O(default_roles_mode) argument.
      - To unset all roles as default, pass an empty list (C([])) and O(default_roles_mode=listed_only).
    type: list
    elements: str
    version_added: '0.6.0'
  roles_mode:
    description:
     - When C(listed_only) (default), makes the user a member of only roles specified in O(roles).
       It will remove the user from all other roles.
     - When C(append), appends roles specified in O(roles) to existing user roles.
     - When C(remove), removes roles specified in O(roles) from user roles.
     - The argument is ignored without O(roles) set.
    type: str
    choices: ['append', 'listed_only', 'remove']
    default: 'listed_only'
    version_added: '0.6.0'
  default_roles_mode:
    description:
     - When C(listed_only) (default), sets only roles specified in O(default_roles) as user default roles.
       It will unset all other roles as default roles.
     - When C(append), appends roles specified in O(default_roles) to existing user default roles.
       default roles instead of unsetting not specified ones.
     - When C(remove), removes roles specified in O(default_roles) from user default roles.
     - Ignored without O(default_roles) set.
    type: str
    choices: ['append', 'listed_only', 'remove']
    default: 'listed_only'
    version_added: '0.6.0'
  user_hosts:
    description:
      - Host restrictions to apply to the user.
      - It's a list of dictionaries, where each dictionary specifies the type of restriction to apply to which hosts or pattern.
    type: list
    elements: dict
    suboptions:
      type:
        description:
          - The method used to validate which hosts that users are allowed to connect from (V(ANY), V(LOCAL), V(IP), V(LIKE), V(NAME), V(REGEXP)).
          - When specified for an existing user the previous host type and hosts will be updated.
          - For more details, see U(https://clickhouse.com/docs/en/sql-reference/statements/create/user).
        type: str
        required: true
      hosts:
        description:
          - A list of hosts or patterns from which the user will be allowed to connect.
          - This is required if O(user_hosts.type) is not V(ANY) or V(LOCAL).
        type: list
        elements: str
        required: false
    version_added: '1.0.0'
  profiles:
    version_added: 2.2.0
    description:
      - Settings profiles that will be applied to role.
      - Can be only used with O(settings) as C(dict).
    type: list
    elements: str
    default: []
'''

EXAMPLES = r'''
- name: Create user granting roles and setting default role
  community.clickhouse.clickhouse_user:
    login_host: localhost
    login_user: alice
    login_db: foo
    login_password: my_password
    name: test_user
    authentication:
      password: qwerty
      type: sha256_password
    roles:
    - accountant
    - manager
    default_roles:
    - accountant

- name: Append the sales role to test_user's roles
  community.clickhouse.clickhouse_user:
    login_host: localhost
    login_user: alice
    login_db: foo
    login_password: my_password
    name: test_user
    roles:
    - sales
    roles_mode: append

- name: Unset all test_user's default roles
  community.clickhouse.clickhouse_user:
    login_host: localhost
    login_user: alice
    login_db: foo
    login_password: my_password
    name: test_user
    default_roles: []

- name: Revoke all roles from test_user
  community.clickhouse.clickhouse_user:
    login_host: localhost
    login_user: alice
    login_db: foo
    login_password: my_password
    name: test_user
    roles: []

- name: If user exists, update password
  community.clickhouse.clickhouse_user:
    login_host: localhost
    login_user: alice
    login_db: foo
    login_password: my_password
    name: test_user
    authentication:
      password: qwerty123
    update_password: always

- name: Update user settings (idempotent - only updates if different)
  community.clickhouse.clickhouse_user:
    login_host: localhost
    login_user: alice
    login_db: foo
    login_password: my_password
    name: test_user
    settings:
      max_memory_usage:
        value: 20000
        writability: readonly
      max_threads:
        value: 8

- name: Update user settings (idempotent - only updates if different)
  community.clickhouse.clickhouse_user:
    login_host: localhost
    login_user: alice
    login_db: foo
    login_password: my_password
    name: test_user
    settings:
      max_memory_usage:
        value: 20G
        min: 10G
        max: 25G
        writability: readonly
      max_threads:
        value: 8
        writability: writable

- name: Create user with specific settings and profile
  community.clickhouse.clickhouse_user:
    login_host: localhost
    login_user: alice
    login_db: foo
    login_password: my_password
    name: test_user
    authentication:
      password: 9e69e7e29351ad837503c44a5971edebc9b7e6d8601c89c284b1b59bf37afa80
      type: sha256_hash
    cluster: test_cluster
    settings:
      max_memory_usage:
        value: 15000
        min: 15000
        max: 16000
        writability: readonly
    profiles:
      - restricted
    state: present

- name: Create a user that can only connect from a specified host
  community.clickhouse.clickhouse_user:
    login_host: localhost
    login_user: alice
    login_db: foo
    login_password: my_password
    name: test_user
    user_hosts:
      - type: NAME
        hosts:
          - 'host1'

- name: Update user host restrictions. Any previous host restrictions will be replaced. (idempotent - only updates if different)
  community.clickhouse.clickhouse_user:
    login_host: localhost
    login_user: alice
    login_db: foo
    login_password: my_password
    name: test_user
    user_hosts:
      - type: LIKE
        hosts:
          - '%.example.com'

- name: Drop user
  community.clickhouse.clickhouse_user:
    login_host: localhost
    login_user: alice
    login_db: foo
    login_password: my_password
    name: test_user
    state: absent

- name: Create user identified with ldap
  community.clickhouse.clickhouse_user:
    name: alice
    state: present
    authentication:
      type: ldap
      server: ad

- name: Create user identified with password
  community.clickhouse.clickhouse_user:
    name: alice
    state: present
    authentication:
      password: my_password

- name: Create user identified with passed hash
  community.clickhouse.clickhouse_user:
    name: alice
    state: present
    authentication:
      type: sha256_hash
      password: my_hash
'''

RETURN = r'''
executed_statements:
  description:
  - Data-modifying executed statements.
  returned: on success
  type: list
  sample: ["CREATE USER test_user IDENTIFIED WITH ***** BY '*****'"]
'''

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.clickhouse.plugins.module_utils.clickhouse import (
    check_clickhouse_driver,
    client_common_argument_spec,
    connect_to_db_via_client,
    execute_query,
    get_main_conn_kwargs,
)
from ansible_collections.community.clickhouse.plugins.module_utils.entity_settings import EntitySettings

executed_statements = []


class ClickHouseUser():
    def __init__(self, module, client, name):
        self.changed = False
        self.module = module
        self.client = client
        self.name = name
        # Set default values, then update
        self.user_exists = False
        self.current_default_roles = []
        self.current_roles = []
        self.current_settings = {}
        self.current_user_hosts = {}
        self.settings = EntitySettings(self.module, self.client, self.name, 'user')

        # Fetch actual values from DB and
        # update the attributes with them
        self.__populate_info()

    def __populate_info(self):
        # Collecting user information
        query = ("SELECT name, storage, auth_type, default_roles_list "
                 "FROM system.users "
                 "WHERE name = '%s'" % self.name)

        result = execute_query(self.module, self.client, query)

        if result != []:
            self.user_exists = True
            self.current_default_roles = result[0][3]

        if self.user_exists:
            self.current_roles = self.__fetch_user_groups()
            self.current_settings = self.__fetch_user_settings()
            self.current_user_hosts = self.__fetch_user_hosts()

    def __fetch_user_groups(self):
        query = ("SELECT granted_role_name FROM system.role_grants "
                 "WHERE user_name = '%s'" % self.name)
        result = execute_query(self.module, self.client, query)
        return [row[0] for row in result]

    def __fetch_user_settings(self):
        """Fetch current user settings from system.settings_profile_elements"""
        query = ("SELECT setting_name, value, min, max, writability, inherit_profile "
                 "FROM system.settings_profile_elements "
                 "WHERE user_name = '%s'" % self.name)
        result = execute_query(self.module, self.client, query)

        # Build a dict of current settings with their full definition
        settings_dict = {}
        for row in result:
            setting_name, value, min_val, max_val, writability, inherit_profile = row

            # Handle PROFILE inheritance separately
            if inherit_profile is not None:
                settings_dict[inherit_profile] = "PROFILE %s" % inherit_profile
                continue

            # Skip if it's not a regular setting (setting_name could be None for profiles)
            if setting_name is None:
                continue

            # Build the setting string as it would appear in ALTER USER
            setting_parts = [setting_name]
            if value is not None:
                setting_parts.extend(["=", str(value)])
            if min_val is not None:
                setting_parts.extend(["MIN", str(min_val)])
            if max_val is not None:
                setting_parts.extend(["MAX", str(max_val)])
            if writability is not None and writability != 'WRITABLE':
                # Map writability enum to SQL keywords
                if writability == 'CONST':
                    setting_parts.append("READONLY")
                elif writability == 'CHANGEABLE_IN_READONLY':
                    setting_parts.append("CHANGEABLE_IN_READONLY")

            settings_dict[setting_name] = " ".join(setting_parts)

        return settings_dict

    def __fetch_user_hosts(self):
        """Fetch current user host restrictions from system.users"""
        query = ("SELECT host_ip, host_names, host_names_regexp, host_names_like FROM system.users "
                 "WHERE name = '%s'" % self.name)
        result = execute_query(self.module, self.client, query)

        user_hosts_dict = {
            'IP': set(result[0][0]),       # HOST ANY is represented by ['::/0'] in host_ip
            'NAME': set(result[0][1]),     # HOST LOCAL is represented by ['localhost'] in host_names
            'REGEXP': set(result[0][2]),
            'LIKE': set(result[0][3]),
        }

        return user_hosts_dict

    def create(self, type_password, password, cluster, user_hosts, settings,
               roles, roles_mode, default_roles, default_roles_mode, authentication, profiles):

        query = "CREATE USER '%s'" % self.name

        ident_clause = self._build_identified_clause(
            auth=authentication,
            legacy_type_password=type_password,
            legacy_password=password,
        )
        query += "%s" % ident_clause

        if user_hosts:
            query += " %s" % self.__build_user_host_clause(user_hosts)

        if cluster:
            query += " ON CLUSTER %s" % cluster

        if settings or profiles:
            if isinstance(settings, list):
                query += " SETTINGS"
                for index, value in enumerate(settings):
                    query += " %s" % value
                    if index < len(settings) - 1:
                        query += ","
            elif isinstance(settings, dict):
                settings_entity = self.settings.compare_and_build_clause(settings, profiles)
                query += settings_entity[1]
            else:
                self.module.fail_json(msg=f"Unexpexted settings passed: {settings}. Supported list or dict.")

        executed_statements.append(query)

        if not self.module.check_mode:
            execute_query(self.module, self.client, query)

        if roles and roles_mode != 'remove':
            self.__grant_roles(roles, cluster)

        if default_roles and default_roles_mode != 'remove':
            self.__set_default_roles(default_roles, cluster)

        return True

    def update(self, update_password, type_password, password, cluster, user_hosts,
               roles, roles_mode, default_roles, default_roles_mode, settings, authentication, profiles):

        if roles is not None:
            self.__update_roles(roles, roles_mode, cluster)

        if default_roles is not None:
            self.__update_default_roles(default_roles, default_roles_mode, cluster)

        self.__update_passwd(update_password, type_password, password, cluster, authentication)

        if user_hosts is not None:
            self.__update_host(user_hosts, cluster)

        if settings is not None or profiles is not None:
            self.__update_settings(settings, cluster, profiles)

        return self.changed

    def drop(self, cluster):
        query = "DROP USER '%s'" % self.name
        if cluster:
            query += " ON CLUSTER %s" % cluster

        executed_statements.append(query)

        if not self.module.check_mode:
            execute_query(self.module, self.client, query)

        return True

    def __update_roles(self, desired_roles, mode, cluster):
        des = set(desired_roles)
        cur = set(self.current_roles)

        if mode == 'remove':
            # Remove only roles already present in current roles
            roles_to_revoke = list(des & cur)
            if roles_to_revoke:
                self.__revoke_roles(roles_to_revoke, cluster)

        elif mode == 'append':
            # Grant only roles from decired that
            # are not already present in current roles
            roles_to_grant = list(des - cur)
            if roles_to_grant:
                self.__grant_roles(roles_to_grant, cluster)

        elif mode == 'listed_only':
            if desired_roles == [] and self.current_roles:
                self.__revoke_roles(self.current_roles, cluster)
            elif desired_roles != []:
                roles_to_grant = list(des - cur)
                roles_to_revoke = list(cur - des)
                if roles_to_grant:
                    self.__grant_roles(roles_to_grant, cluster)
                if roles_to_revoke:
                    self.__revoke_roles(roles_to_revoke, cluster)

    def __update_default_roles(self, desired_def_roles, mode, cluster):
        des = set(desired_def_roles)
        cur = set(self.current_default_roles)

        if mode == 'remove' and des & cur:
            if desired_def_roles != []:
                # In this case, "desired" means "desired to get removed"
                self.__set_default_roles(list(cur - des), cluster)

        elif mode == 'append':
            if desired_def_roles != [] and des != cur:
                self.__set_default_roles(list(cur.union(des)), cluster)

        elif mode == 'listed_only':
            if desired_def_roles == [] and self.current_roles:
                self.__unset_default_roles()

            elif desired_def_roles != [] and des != cur:
                self.__set_default_roles(desired_def_roles, cluster)

    def __update_passwd(self, update_password, type_pwd, pwd, cluster, authentication):
        if update_password == 'on_create':
            return False or self.changed

        # If update_password is always
        # TODO: When ClickHouse will allow to retrieve password hashes,
        # make this idempotent, i.e. execute this only if the passwords don't match

        query = "ALTER USER '%s'" % self.name

        if cluster:
            query += " ON CLUSTER %s" % cluster

        ident_clause = self._build_identified_clause(
            auth=authentication,
            legacy_type_password=type_pwd,
            legacy_password=pwd,
        )

        query += "%s" % ident_clause

        executed_statements.append(query)

        if not self.module.check_mode:
            execute_query(self.module, self.client, query)

        self.changed = True

    def __update_host(self, user_hosts, cluster):
        des = self.__get_desired_user_hosts(user_hosts)
        cur = self.current_user_hosts

        if des == cur:
            return

        query = "ALTER USER '%s' %s" % (self.name, self.__build_user_host_clause(user_hosts))

        if cluster:
            query += " ON CLUSTER %s" % cluster

        executed_statements.append(query)

        if not self.module.check_mode:
            execute_query(self.module, self.client, query)

        self.changed = True

    def __get_desired_user_hosts(self, user_hosts):
        desired_hosts = {
            'IP': set(),
            'NAME': set(),
            'REGEXP': set(),
            'LIKE': set(),
        }
        for host_restriction in user_hosts:
            host_type = host_restriction['type'].upper()
            if host_type == 'ANY':
                # ANY overrides all other restrictions
                return {
                    'IP': set(['::/0']),
                    'NAME': set(),
                    'REGEXP': set(),
                    'LIKE': set(),
                }
            elif host_type.upper() == 'LOCAL':
                desired_hosts['NAME'].add('localhost')
            else:
                hosts = host_restriction.get('hosts', [])
                desired_hosts[host_type].update(hosts)

        return desired_hosts

    def __build_user_host_clause(self, user_hosts):
        # If ANY is specified among host restrictions, it overrides all others
        if 'ANY' in [hr['type'].upper() for hr in user_hosts]:
            return "HOST ANY"

        clauses = []
        for host_restriction in user_hosts:
            host_type = host_restriction['type']

            if host_type.upper() == 'LOCAL':
                clauses.append("HOST %s" % host_type)
            else:
                hosts = host_restriction.get('hosts')
                clauses.append("HOST %s %s" % (host_type, ', '.join("'{0}'".format(h) for h in hosts)))

        return ' '.join(clauses)

    def __grant_roles(self, roles_to_set, cluster):
        query = "GRANT %s TO '%s'" % (', '.join(roles_to_set), self.name)
        executed_statements.append(query)

        if cluster:
            query += " ON CLUSTER %s" % cluster

        if not self.module.check_mode:
            execute_query(self.module, self.client, query)

        self.changed = True

    def __revoke_roles(self, roles_to_revoke, cluster):
        query = "REVOKE %s FROM '%s'" % (', '.join(roles_to_revoke), self.name)
        executed_statements.append(query)

        if cluster:
            query += " ON CLUSTER %s" % cluster

        if not self.module.check_mode:
            execute_query(self.module, self.client, query)

        self.changed = True

    def __set_default_roles(self, roles_to_set, cluster):
        self.current_roles = self.__fetch_user_groups()
        for role in roles_to_set:
            if role not in self.current_roles and role not in self.module.params["roles"]:
                self.module.fail_json("User %s is not in %s role. Grant it explicitly first." % (self.name, role))

        query = "ALTER USER '%s' DEFAULT ROLE %s" % (self.name, ', '.join(roles_to_set))
        if cluster:
            query += " ON CLUSTER %s" % cluster
        executed_statements.append(query)

        if not self.module.check_mode:
            execute_query(self.module, self.client, query)

        self.changed = True

    def __unset_default_roles(self):
        query = "SET DEFAULT ROLE NONE TO '%s'" % self.name
        executed_statements.append(query)

        if not self.module.check_mode:
            execute_query(self.module, self.client, query)

        self.changed = True

    def __update_settings(self, settings, cluster, profiles):
        """Update user settings idempotently by comparing with current settings"""

        if isinstance(settings, list):
            # Parse desired settings into a comparable format
            desired_settings = {}
            desired_profiles = []

            for setting in settings:
                setting_upper = setting.upper()
                # Handle PROFILE separately as it's not a regular setting
                if 'PROFILE' in setting_upper:
                    # Extract profile name (handle both PROFILE 'name' and PROFILE name)
                    profile_part = setting.split(None, 1)[1].strip().strip("'\"")
                    desired_profiles.append(profile_part)
                else:
                    # Extract setting name (first word before = or space)
                    setting_name = setting.split()[0].split('=')[0].strip()
                    # Normalize the setting string for comparison
                    normalized = ' '.join(setting.split())
                    desired_settings[setting_name] = normalized

            # Compare current with desired
            needs_update = False

            # Check if any setting values differ
            for setting_name, desired_def in desired_settings.items():
                current_def = self.current_settings.get(setting_name, '')
                # Normalize both for comparison (case-insensitive, whitespace-normalized)
                current_normalized = ' '.join(current_def.upper().split())
                desired_normalized = ' '.join(desired_def.upper().split())

                if current_normalized != desired_normalized:
                    needs_update = True
                    break

            # Check if there are settings to remove (current has settings not in desired)
            if not needs_update:
                for setting_name in self.current_settings:
                    if setting_name not in desired_settings:
                        # There's a setting currently applied that's not in desired
                        # We need to reapply to remove it
                        needs_update = True
                        break

            # Only update if settings actually differ
            if not needs_update:
                return

            # Build the ALTER USER query
            query = "ALTER USER '%s' SETTINGS" % self.name
            for index, value in enumerate(settings):
                query += " %s" % value
                if index < len(settings) - 1:
                    query += ","

            if cluster:
                query += " ON CLUSTER %s" % cluster
        elif isinstance(settings, dict) or profiles:
            query = "ALTER USER '%s'" % self.name
            changed, settings_entity, changes = self.settings.compare_and_build_clause(settings, profiles)
            if not changed:
                return
            query += settings_entity
            if cluster:
                query += " ON CLUSTER %s" % cluster
        else:
            return
        executed_statements.append(query)

        if not self.module.check_mode:
            execute_query(self.module, self.client, query)

        self.changed = True

    def _build_identified_clause(self, auth=None, legacy_type_password=None, legacy_password=None):
        if auth and isinstance(auth, dict):
            atype = auth.get('type', '').lower().strip()

            if atype == "not_identified":
                return " NOT IDENTIFIED"

            if atype == 'no_password':
                return " IDENTIFIED WITH NO_PASSWORD"

            if atype == 'ldap':
                server = auth.get('server')
                if not server:
                    self.module.fail_json(msg="authentication.type=ldap requires 'server'")
                return f" IDENTIFIED WITH ldap SERVER '{server}'"

            password = auth.get('password') or legacy_password
            if password is None:
                self.module.fail_json(msg=f"type {atype} requires password")

            if atype in ('plaintext_password', 'sha256_password', 'double_sha1_password', 'bcrypt_password'):
                return f" IDENTIFIED WITH {atype} BY '{password}'"

            # Separate so maybe in future add salt here.
            if atype in ('sha256_hash', 'double_sha1_hash', 'bcrypt_hash'):
                return f" IDENTIFIED WITH {atype} BY '{password}'"

            self.module.fail_json(msg=f"type {atype} not implemented")

        # Backward compatibility
        if legacy_password is not None:
            return f" IDENTIFIED WITH {legacy_type_password} BY '{legacy_password}'"
        return ""


def main():
    argument_spec = client_common_argument_spec()
    argument_spec.update(
        state=dict(type='str', choices=['present', 'absent'], default='present'),
        name=dict(type='str', required=True),
        password=dict(
            type='str',
            default=None,
            no_log=True,
            removed_in_version='3.0.0',
            removed_from_collection='community.clickhouse'
        ),
        type_password=dict(
            type='str',
            default='sha256_password',
            no_log=False,
            removed_in_version='3.0.0',
            removed_from_collection='community.clickhouse'
        ),
        authentication=dict(
            type='dict',
            default=None,
            options=dict(
                type=dict(
                    type='str',
                    default='sha256_password',
                    choices=[
                        'not_identified',
                        'no_password',
                        'plaintext_password',
                        'sha256_password',
                        'sha256_hash',
                        'double_sha1_password',
                        'double_sha1_hash',
                        'bcrypt_password',
                        'bcrypt_hash',
                        'ldap'
                    ]
                ),
                server=dict(type='str'),
                password=dict(type='str', no_log=True),
            ),
        ),
        cluster=dict(type='str', default=None),
        update_password=dict(
            type='str', choices=['always', 'on_create'],
            default='on_create', no_log=False
        ),
        user_hosts=dict(type='list', elements='dict'),
        settings=dict(type='raw'),
        profiles=dict(type='list', elements='str', default=[]),
        roles=dict(type='list', elements='str', default=None),
        default_roles=dict(type='list', elements='str', default=None),
        roles_mode=dict(type='str', choices=['listed_only', 'append', 'remove'],
                        default='listed_only'),
        default_roles_mode=dict(type='str', choices=['listed_only', 'append', 'remove'],
                                default='listed_only')
    )

    # Instantiate an object of module class
    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        mutually_exclusive=[['password', 'authentication']],
    )

    # Assign passed options to variables
    client_kwargs = module.params['client_kwargs']
    # The reason why these arguments are separate from client_kwargs
    # is that we need to protect some sensitive data like passwords passed
    # to the module from logging (see the arguments above with no_log=True);
    # Such data must be passed as module arguments (not nested deep in values).
    main_conn_kwargs = get_main_conn_kwargs(module)
    state = module.params['state']
    name = module.params['name']
    password = module.params["password"]
    type_password = module.params["type_password"]
    cluster = module.params['cluster']
    update_password = module.params['update_password']
    user_hosts = module.params['user_hosts']
    settings = module.params['settings']
    profiles = module.params['profiles']
    roles = module.params['roles']
    roles_mode = module.params['roles_mode']
    default_roles = module.params['default_roles']
    default_roles_mode = module.params['default_roles_mode']
    authentication = module.params['authentication']

    # Will fail if no driver informing the user
    check_clickhouse_driver(module)

    # Connect to DB
    client = connect_to_db_via_client(module, main_conn_kwargs, client_kwargs)

    # Do the job
    changed = False
    user = ClickHouseUser(module, client, name)
    if state == 'present':
        if not user.user_exists:
            changed = user.create(type_password, password, cluster, user_hosts, settings,
                                  roles, roles_mode, default_roles, default_roles_mode, authentication, profiles)
        else:
            # If user exists
            changed = user.update(update_password, type_password, password, cluster, user_hosts,
                                  roles, roles_mode, default_roles, default_roles_mode, settings, authentication, profiles)
    else:
        # If state is absent
        if user.user_exists:
            changed = user.drop(cluster)

    # Close connection
    client.disconnect_connection()

    # Users will get this in JSON output after execution
    module.exit_json(changed=changed, executed_statements=executed_statements)


if __name__ == '__main__':
    main()
